from datetime import datetime as datetime
from typing import Tuple

__all__ = ["calc_time"]


def calc_time(time0: datetime, time1: datetime) -> Tuple[int, int, int, int, int]:
    """Calculate the time delta between time0 and time1.

    Calculates the time delta between time0 and time1 and
    returns the result as a tuple of integers representing the number of
    months, days, hours, minutes and seconds
    the time delta constitutes.

    Parameters
    ----------
    time0 : datetime
        start time
    time1 : datetime
        end time

    Returns
    -------
    Tuple[int, int, int, int, int]
        tuple of integers representing the number of
        months, days, hours, minutes and seconds
        the time delta constitutes.
    """
    _second: int = (time1 - time0).seconds
    _minute: float = _second / 60
    _hour: float = _minute / 60
    _day: float = _hour / 24
    _month: float = _day / 30
    second: int = int(_second % 60)
    minute: int = int(_minute % 60)
    hour: int = int(_hour % 24)
    day: int = int(_day % 30)
    month: int = int(_month)

    return month, day, hour, minute, second



import re
from collections import OrderedDict
from typing import Any, List, MutableMapping, Set, Union


def find_key(dict: MutableMapping[Any, Any], pattern: str) -> Union[str, None]:
    """Find the first key in dict that matches the given pattern."""
    try:
        return [k for k in dict.keys() if re.search(pattern, k)][0]
    except Exception:
        return None


def find_keys(dict: MutableMapping[Any, Any], pattern: str) -> Union[List[str], None]:
    """Find all keys in dict that match the given pattern."""
    try:
        return [k for k in dict.keys() if re.search(pattern, k)]
    except Exception:
        return None


def find_type_identifier_in_keys(dict: MutableMapping[Any, Any]) -> Union[str, None]:
    """Find the first key name in dict that contains one of the following type identifier strings:
    [Integer|Real|Boolean|Enumeration|String|Unknown].
    """
    key_list: List[str] = ["Integer", "Real", "Boolean", "Enumeration", "String", "Unkown"]
    type_identifier: List[str] = []
    for key in dict:
        key_without_index = re.sub(r"^\d{6}_", "", key)

        if key_without_index in key_list:
            type_identifier.append(key_without_index)

    return type_identifier[0] if type_identifier else None


def shrink_dict(dict: MutableMapping[Any, Any], unique_key: Union[List[str], None] = None) -> MutableMapping[Any, Any]:
    """Identify doubled entries in the passed in dict and return a new dict with doubled entries removed."""
    _unique_key: List[str] = unique_key or []
    unique_keys_string: str = "['" + "']['".join(_unique_key) + "']"
    # sort an ordered dict for attribute (child) where the dict is to make unique for
    eval_string: str = f"sorted(dict.items(), key=lambda x: str(x[1]{unique_keys_string}))"

    # Identify doublettes and collect them for subsequent removal
    seen: Set[Any] = set([])
    remove_key: List[Any] = []

    # value is necessary here as it is used in the eval statements below. Do not delete it.
    for key, value in OrderedDict(eval(eval_string)).items():  # noqa: B007
        proove_value = eval(f"value{unique_keys_string}")
        if proove_value in seen:
            remove_key.append(key)
        else:
            seen.add(eval(f"value{unique_keys_string}"))

    return {key: dict[key] for key in dict.keys() if key not in remove_key}



#!/usr/bin/env python
# coding: utf-8

import argparse
import logging
from pathlib import Path
from typing import Union

from ospx import OspCaseBuilder
from ospx.utils.logging import configure_logging

logger = logging.getLogger(__name__)


def _argparser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="ospCaseBuilder",
        usage="%(prog)s caseDict [options [args]]",
        epilog="_________________ospCaseBuilder___________________",
        prefix_chars="-",
        add_help=True,
        description="Builds the OSP-specific configuration files needed to run an OSP (co-)simulation case.",
    )

    _ = parser.add_argument(
        "caseDict",
        metavar="caseDict",
        type=str,
        nargs="?",
        help="name of the dict file containing the OSP simulation case configuration.",
    )

    _ = parser.add_argument(
        "--clean",
        action="store_true",
        help=(
            "cleans up working directory and deletes any existing ospx files, e.g. modelDescription.xml .fmu .csv etc."
        ),
        default=False,
        required=False,
    )

    _ = parser.add_argument(
        "-i",
        "--inspect",
        action="store_true",
        help=("inspect mode: reads all properties from the FMUs but does not actually create the OSP case files."),
        default=False,
        required=False,
    )

    _ = parser.add_argument(
        "-g",
        "--graph",
        action="store_true",
        help="creates a dependency graph image using graphviz",
        default=False,
        required=False,
    )

    console_verbosity = parser.add_mutually_exclusive_group(required=False)

    _ = console_verbosity.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help=("console output will be quiet."),
        default=False,
    )

    _ = console_verbosity.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help=("console output will be verbose."),
        default=False,
    )

    _ = parser.add_argument(
        "--log",
        action="store",
        type=str,
        help="name of log file. If specified, this will activate logging to file.",
        default=None,
        required=False,
    )

    _ = parser.add_argument(
        "--log-level",
        action="store",
        type=str,
        help="log level applied to logging to file.",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="WARNING",
        required=False,
    )

    return parser


def main():
    """Entry point for console script as configured in setup.cfg.

    Runs the command line interface and parses arguments and options entered on the console.
    """
    parser = _argparser()
    args = parser.parse_args()

    # Configure Logging
    # ..to console
    log_level_console: str = "INFO"
    if any([args.quiet, args.verbose]):
        log_level_console = "ERROR" if args.quiet else log_level_console
        log_level_console = "DEBUG" if args.verbose else log_level_console

    # ..to file
    log_file: Union[Path, None] = Path(args.log) if args.log else None
    log_level_file: str = args.log_level
    configure_logging(log_level_console, log_file, log_level_file)

    inspect: bool = args.inspect
    graph: bool = args.graph
    clean: bool = args.clean

    case_dict_file: Path = Path(args.caseDict)

    # Check whether case dict file exists
    if not case_dict_file.exists():
        logger.error(f"ospCaseBuilder.py: File {case_dict_file} not found.")
        return

    logger.info(
        f"Start ospCaseBuilder.py with following arguments:\n"
        f"\t case_dict_file: \t\t{case_dict_file}\n"
        f"\t inspect: \t\t\t\t{inspect}\n"
        f"\t graph: \t\t\t{graph}\n"
        f"\t clean: \t\t\t{clean}\n"
    )

    # Invoke API
    OspCaseBuilder.build(
        case_dict_file=case_dict_file,
        inspect=inspect,
        graph=graph,
        clean=clean,
    )


if __name__ == "__main__":
    main()



import logging
from collections.abc import Iterable
from typing import Any, List, Sequence, Union

from dictIO import Formatter, Parser

__ALL__ = ["ScalarVariable", "get_fmi_data_type"]

logger = logging.getLogger(__name__)


class ScalarVariable:
    """fmi 2.0 ScalarVariable.

    See https://github.com/modelica/fmi-standard/blob/v2.0.x/schema/fmi2ScalarVariable.xsd
    """

    def __init__(
        self,
        name: str,
        data_type: Union[str, None] = None,
        causality: Union[str, None] = None,
        variability: Union[str, None] = None,
        start: Union[int, float, bool, str, None] = None,
        value_reference: int = 0,
        description: Union[str, None] = None,
        quantity: Union[str, None] = None,
        unit: Union[str, None] = None,
        display_unit: Union[str, None] = None,
    ):
        # Attributes
        self.name: str
        self._data_type: Union[str, None] = None
        self._causality: str = "local"
        self._variability: Union[str, None] = None
        self._start: Union[int, float, bool, str, None] = None
        self.value_reference: int = 0
        self.description: Union[str, None] = None
        self.quantity: Union[str, None] = None
        self.unit: Union[str, None] = None
        self.display_unit: Union[str, None] = None
        # Initialization
        self.name = name
        if data_type:
            self.data_type = data_type
        if causality:
            self.causality = causality
        if variability:
            self.variability = variability
        if start:
            self.start = start
        self.value_reference = value_reference
        self.description = description
        self.quantity = quantity
        self.unit = unit
        self.display_unit = display_unit

    @property
    def data_type(self) -> Union[str, None]:
        """Returns the FMI data type of the scalar Variable."""
        return self._data_type

    @data_type.setter
    def data_type(self, type: str):
        """Set the FMI data type of the scalar Variable.

        Valid values are:
            "Real"
            "Integer"
            "Boolean"
            "String"
            "Enumeration"
        """
        valid_types: list[str] = [
            "Real",
            "Integer",
            "Boolean",
            "String",
            "Enumeration",
        ]
        if type not in valid_types:
            logger.error(f"variable {self.name}: value for data_type '{type}' is invalid.")
            return
        self._data_type = type
        return

    @property
    def causality(self) -> str:
        """Returns the causality of the scalar Variable."""
        return self._causality

    @causality.setter
    def causality(self, value: str):
        """Set the causality of the scalar Variable.

        Valid values are:
            "parameter"
            "calculatedParameter"
            "input"
            "output"
            "local"
            "independent"
            "structuralParameter"
        """
        valid_values: list[str] = [
            "parameter",
            "calculatedParameter",
            "input",
            "output",
            "local",
            "independent",
            "structuralParameter",
        ]
        if value not in valid_values:
            logger.error(f"variable {self.name}: causality value '{value}' is invalid.")
            return
        self._causality = value
        return

    @property
    def variability(self) -> Union[str, None]:
        """Returns the variability of the scalar Variable."""
        return self._variability

    @variability.setter
    def variability(self, value: str):
        """Set the variability of the scalar Variable.

        Valid values are:
            "constant"
            "fixed"
            "tunable"
            "discrete"
            "continuous"
        """
        valid_values: list[str] = [
            "constant",
            "fixed",
            "tunable",
            "discrete",
            "continuous",
        ]
        if value not in valid_values:
            logger.error(f"variable {self.name}: value for variability '{value}' is invalid.")
            return
        self._variability = value
        return

    @property
    def start(self) -> Union[int, float, bool, str, None]:
        """Returns the start value (initial value) of the scalar Variable."""
        return self._start

    @start.setter
    def start(self, value: Union[int, float, bool, str, None]):
        """Set the start value (initial value) of the scalar Variable."""
        if value is None:
            logger.error(f"variable {self.name}: start value shall be set to 'None', but 'None' is invalid for start.")
            return
        if self.data_type:
            # make sure the data type of the new value does either match or gets casted to the data_type defined for the variable
            new_value_data_type = get_fmi_data_type(value)
            if new_value_data_type == self.data_type:
                self._start = value
            else:
                casted_value = _cast_to_fmi_data_type(value, self.data_type)
                if casted_value is not None and not isinstance(casted_value, Sequence):
                    self._start = casted_value
                elif casted_value is None:
                    logger.error(
                        f"variable {self.name}: start shall be set to 'None', but 'None' is invalid for start."
                    )
                    return
                else:
                    logger.error(
                        f"variable {self.name}: start shall be set to {casted_value}, but fmi data type 'Enumeration' is invalid for start."
                    )
                    return
        else:
            self._start = value
            self.data_type = get_fmi_data_type(self.start)


def get_fmi_data_type(arg: Any) -> str:
    r"""Return the fmi 2.0 data type corresponding to Python type of the passed in argument.

    See https://github.com/modelica/fmi-standard/blob/v2.0.x/schema/fmi2Type.xsd

    Parameters
    ----------
    arg : Any
        The argument for which the fmi 2.0 data type shall be determined

    Returns
    -------
    str
        The fmi 2.0 data type, returned as string literal.\n
        valid fmi 2.0 data types are 'Integer', 'Real', 'Boolean', 'String' and 'Enumeration'
    """

    if isinstance(arg, int):
        return "Integer"
    elif isinstance(arg, float):
        return "Real"
    elif isinstance(arg, bool):
        return "Boolean"
    # not regarding the content, sequence is always returned if not int or float, e.g. string.
    # requires a solution, if xs:enumeration is required.
    # elif isinstance(arg, Sequence):
    #    return 'Enumeration'
    else:
        return "String"


def _cast_to_fmi_data_type(
    arg: Union[int, float, bool, str, Sequence[Any]], fmi_data_type: str
) -> Union[int, float, bool, str, List[Any], None]:
    r"""Casts the passed in argument to a Python data type that matches the requested fmi data type.

    Parameters
    ----------
    arg : Union[int, float, bool, str, Sequence[Any]]
        The argument to be casted
    fmi_data_type : str
        The fmi data type the argument shall be casted to.\n
        valid fmi 2.0 data types are 'Integer', 'Real', 'Boolean', 'String' and 'Enumeration'

    Returns
    -------
    Union[int, float, bool, str, List[Any], None]
        The casted value (in a Python data type that matches the requested fmi data type)
    """
    if fmi_data_type in {"Integer", "Real", "Boolean"}:
        if isinstance(arg, Sequence):
            logger.warning(
                f"_cast_to_fmi_data_type(): argument {arg} of type List/Tuple/Sequence cannot be casted to fmi data type {fmi_data_type}"
            )
            return None
        # parse if arg is string
        parsed_value: Union[int, float, bool]
        parsed_value = Parser().parse_type(arg) if isinstance(arg, str) else arg
        # cast to int / float / bool
        if fmi_data_type == "Integer":
            return int(parsed_value)
        elif fmi_data_type == "Real":
            return float(parsed_value)
        else:
            return bool(parsed_value)
    elif fmi_data_type == "String":
        # format as string
        return Formatter().format_dict(arg) if isinstance(arg, Sequence) else Formatter().format_type(arg)
    elif fmi_data_type == "Enumeration":
        # cast to list
        return list(arg) if isinstance(arg, Iterable) else [arg]
    else:
        return None



#!/usr/bin/env python
# coding: utf-8

import argparse
import logging
import os
import re
import shutil
from argparse import ArgumentParser
from pathlib import Path
from time import sleep
from typing import List, Union

from ospx.utils.logging import configure_logging
from ospx.watch.watchCosim import CosimWatcher

logger = logging.getLogger(__name__)


def _argparser() -> argparse.ArgumentParser:
    parser = ArgumentParser(
        prog="watchCosim",
        usage="%(prog)s watchDict [options [args]]",
        epilog="_________________watchCosim___________________",
        prefix_chars="-",
        add_help=True,
        description="Continuously watches the progress of cosim, and finally saves results as a pandas dataframe.",
    )

    _ = parser.add_argument(
        "watchDict",
        metavar="watchDict",
        type=str,
        help="name of the dict file containing the watch configuration (will also be part of the result file names).",
    )

    _ = parser.add_argument(
        "-c",
        "--converge",
        action="store_true",
        help="watch convergence progress, finally --dump (reading watchDict and .csv, plotting convergence until no changes happen for 5s to any .csv)",
        default=False,
        required=False,
    )

    _ = parser.add_argument(
        "-d",
        "--dump",
        action="store_true",
        help="dump data (reading watchDict and .csv, creating results/{dataFrameDump, resultsDict})",
        default=False,
        required=False,
    )

    _ = parser.add_argument(
        "-p",
        "--plot",
        action="store_true",
        help="plot data including --dump (reading watchDict and .csv, creating results/SIMULATIONNAME.png)",
        default=False,
        required=False,
    )

    console_verbosity = parser.add_mutually_exclusive_group(required=False)

    _ = console_verbosity.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help=("console output will be quiet."),
        default=False,
    )

    _ = console_verbosity.add_argument(
        "-t",
        "--timeline",
        action="store_true",
        help=("write all timeline data into resultDict"),
        default=False,
    )

    _ = console_verbosity.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help=("console output will be verbose."),
        default=False,
    )

    _ = parser.add_argument(
        "--latest",
        action="store",
        type=int,
        help="specify the interval of latest n timesteps to be taken into account",
        default=0,
        required=False,
    )

    _ = parser.add_argument(
        "--log",
        action="store",
        type=str,
        help="name of log file. If specified, this will activate logging to file.",
        default=None,
        required=False,
    )

    _ = parser.add_argument(
        "--log-level",
        action="store",
        type=str,
        help="log level applied to logging to file.",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="WARNING",
        required=False,
    )
    _ = parser.add_argument(
        "--scale",
        action="store",
        type=int,
        help="scale image by factor",
        default=1,
        required=False,
    )
    _ = parser.add_argument(
        "--skip",
        action="store",
        type=int,
        help="skip the first n timesteps",
        default=0,
        required=False,
    )

    return parser


def main():
    """Entry point for console script as configured in setup.cfg.

    Runs the command line interface and parses arguments and options entered on the console.
    """

    parser = _argparser()
    args = parser.parse_args()

    # Configure Logging
    # ..to console
    log_level_console: str = "WARNING"
    if any([args.quiet, args.verbose]):
        log_level_console = "ERROR" if args.quiet else log_level_console
        log_level_console = "DEBUG" if args.verbose else log_level_console
    # ..to file
    log_file: Union[Path, None] = Path(args.log) if args.log else None
    log_level_file: str = args.log_level
    configure_logging(log_level_console, log_file, log_level_file)

    watch_dict_file_name: str = args.watchDict
    converge: bool = args.converge
    plot: bool = args.plot
    dump: bool = args.dump
    skip: int = args.skip
    latest: int = args.latest
    scale_factor = float(args.scale)
    timeline_data: bool = args.timeline

    if not converge and not plot and not dump:
        logger.error("give at least one option what to do: --converge, --plot or --dump")
        parser.print_help()
        exit(0)

    # Dispatch to _main(), which takes care of processing the arguments and invoking the API.
    _main(
        watch_dict_file_name=watch_dict_file_name,
        converge=converge,
        plot=plot,
        dump=dump,
        skip_values=skip,
        latest_values=latest,
        scale_factor=scale_factor,
        timeline_data=timeline_data,
    )


def _main(
    watch_dict_file_name: str,
    converge: bool = False,
    plot: bool = False,
    dump: bool = False,
    skip_values: int = 0,
    latest_values: int = 0,
    scale_factor: float = 1,
    timeline_data: bool = False,
):
    """Entry point for unit tests.

    Processes the arguments parsed by main() on the console and invokes the API.
    """

    watch_dict_file = Path(watch_dict_file_name)

    if not watch_dict_file.is_file():
        logger.error(f"file {watch_dict_file} not found.")
        return

    csv_files: List[Path] = []
    wait_counter: int = 0
    while wait_counter < 5:
        csv_files = list(Path(".").glob("*.csv"))
        if csv_files:
            break
        if wait_counter == 0:
            logger.warning("waiting for csv files..")
        else:
            logger.info("waiting for csv files.")
        wait_counter += 1
        sleep(1)

    if not csv_files:
        logger.error("no csv files found.")
        return

    csv_file_names: List[str] = sorted(file.name for file in csv_files)

    # From the csv file names, identify all data sources for which csv files have been written,
    # and save them as set.
    data_source_names = {re.sub(r"_\d{8}_\d{6}_\d{6}.*$", "", n) for n in csv_file_names}

    # For each identified component name:
    # 1. Find all csv files that match the component name
    # 2. From these, collect only the latest (newest) version
    latest_csv_file_names = [
        sorted(
            [file_name for file_name in csv_file_names if re.match(data_source_name, file_name)],
            key=os.path.getmtime,
        )[-1]
        for data_source_name in data_source_names
    ]

    watcher = CosimWatcher(latest_csv_file_names, skip_values, latest_values, scale_factor, timeline_data)
    watcher.read_watch_dict(watch_dict_file_name)

    Path(watcher.results_dir).mkdir(parents=True, exist_ok=True)

    if converge:
        watcher.plot(converge=True)

    elif plot:
        watcher.plot()

    if dump:
        watcher.dump()

        # finally: move annoying csv files to results
        # after simplifying watchCosim or splitting off, it can be considered to move csv files in advance
        # (not suitable for -c option!)
        for file in csv_files:
            shutil.move(file, "results")


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import List

import pytest

from ospx.watch.watchCosim import CosimWatcher


def test_file_not_found_exception():
    # Prepare
    source_file: Path = Path("this_file_does_not_exist")
    csv_file_names: List[str] = []
    skip_values: int = 0
    latest_values: int = 0
    watcher = CosimWatcher(
        csv_file_names,
        skip_values,
        latest_values,
        scale_factor=1.0,
        timeline_data=False,
    )
    # Execute and Assert
    with pytest.raises(FileNotFoundError):
        watcher.read_watch_dict(source_file)


def test_read_watch_dict():
    # Prepare
    source_file = Path("test_watchDict")
    csv_file_names = [
        "test_result_file_difference.csv",
        "test_result_file_divident.csv",
        "test_result_file_minuend.csv",
        "test_result_file_quotient.csv",
        "test_result_file_subtrahend.csv",
    ]
    skip_values = 0
    latest_values = 0
    watcher = CosimWatcher(
        csv_file_names,
        skip_values,
        latest_values,
        scale_factor=1.0,
        timeline_data=False,
    )
    # Execute
    watcher.read_watch_dict(source_file)
    # Assert



import logging
from dataclasses import dataclass, field
from typing import Union

__ALL__ = ["Unit", "BaseUnit", "DisplayUnit"]

logger = logging.getLogger(__name__)


@dataclass()
class BaseUnit:
    r"""fmi 2.0 BaseUnit.

    Unit definition with reference to SI base units \n
    base unit value = factor * unit value + offset \n
    See https://github.com/modelica/fmi-standard/blob/v2.0.x/schema/fmi2Unit.xsd
    """

    kg: int = 0
    m: int = 0
    s: int = 0
    A: int = 0
    K: int = 0
    mol: int = 0
    cd: int = 0
    rad: int = 0
    factor: float = 1.0
    offset: float = 0.0


@dataclass()
class DisplayUnit:
    r"""fmi 2.0 DisplayUnit.

    display unit value = factor * unit value + offset \n
    See https://github.com/modelica/fmi-standard/blob/v2.0.x/schema/fmi2Unit.xsd
    """

    # Name of DisplayUnit element, e.g. <Unit name="rad"/>, <DisplayUnit name="deg" factor="57.29..."/>.
    # Name must be unique with respect to all other DisplayUnits defined inside a Unit element
    # (in contrast, multiple Unit elements may have DisplayUnits with the same name).
    name: str = field(default_factory=lambda: "-")
    # display_unit value = factor * unit value + offset
    factor: float = 1.0
    offset: float = 0.0


@dataclass()
class Unit:
    """fmi 2.0 Unit Definition.

    See https://github.com/modelica/fmi-standard/blob/v2.0.x/schema/fmi2Unit.xsd
    """

    # Name of unit, e.g. "N.m", "Nm",  "%/s".
    # "name" must be unique with respect to all other unit elements inside the UnitDefinitions section.
    name: str = field(default_factory=lambda: "-")
    base_unit: Union[BaseUnit, None] = None
    display_unit: DisplayUnit = field(default_factory=lambda: DisplayUnit())



import logging
import os
from pathlib import Path
from typing import Union

from dictIO import CppDict, DictReader

from ospx import Graph, OspSimulationCase

__ALL__ = ["OspCaseBuilder"]

logger = logging.getLogger(__name__)


class OspCaseBuilder:
    """Builder for OSP-specific configuration files needed to run an OSP (co-)simulation case."""

    def __init__(self):
        return

    @staticmethod
    def build(
        case_dict_file: Union[str, os.PathLike[str]],
        inspect: bool = False,
        graph: bool = False,
        clean: bool = False,
    ):
        """Build the OSP-specific configuration files needed to run an OSP (co-)simulation case.

        Builds following files:
            - OspSystemStructure.xml
            - SystemStructure.ssd
            - Plot.json
            - statisticsDict
            - watchDict

        Parameters
        ----------
        case_dict_file : Union[str, os.PathLike[str]]
            caseDict file. Contains all case-specific information OspCaseBuilder needs to generate the OSP files.
        inspect : bool, optional
            inspect mode. If True, build() reads all properties from the FMUs but does not actually create the OSP case files, by default False
        graph : bool, optional
            if True, creates a dependency graph image using graphviz, by default False
        clean : bool, optional
            if True, cleans up case folder and deletes any formerly created ospx files, e.g. OspSystemStructure.xml .fmu .csv etc.

        Raises
        ------
        FileNotFoundError
            if case_dict_file does not exist
        """

        # Make sure source_file argument is of type Path. If not, cast it to Path type.
        case_dict_file = case_dict_file if isinstance(case_dict_file, Path) else Path(case_dict_file)
        if not case_dict_file.exists():
            logger.error(f"OspCaseBuilder: File {case_dict_file} not found.")
            raise FileNotFoundError(case_dict_file)

        if clean:
            case_folder: Path = case_dict_file.resolve().parent
            _clean_case_folder(case_folder)

        logger.info(f"reading {case_dict_file}")  # 0

        case_dict: CppDict = DictReader.read(case_dict_file, comments=False)

        case = OspSimulationCase(case_dict)
        try:
            case.setup()
        except Exception as e:
            logger.exception(e)
            return

        if inspect:
            # inspect and return
            case._inspect()  # pyright: ignore
            return

        # case.write_osp_model_description_xmls()
        case.write_osp_system_structure_xml()
        case.write_system_structure_ssd()

        if "postProcessing" in case_dict.keys():
            case._write_plot_config_json()  # pyright: ignore

        case.write_statistics_dict()

        if graph:
            Graph.generate_dependency_graph(case)

        case.write_watch_dict()

        return


def _clean_case_folder(case_folder: Path):
    """Clean up the case folder and deletes any existing ospx files, e.g. modelDescription.xml .fmu .csv etc."""
    import re
    from shutil import rmtree

    # specify all files to be deleted (or comment-in / comment-out as needed)
    case_builder_result_files = [
        "*.csv",
        "*.out",
        "*.xml",
        "*.ssd",
        "*.fmu",
        "*callGraph",
        "*.pdf",
        "*.png",  # 'protect results/*.png'
        "watchDict",
        "statisticsDict",  # 'results',
        "zip",
    ]
    except_list = ["src", "^test_", "_OspModelDescription.xml"]
    except_pattern = "(" + "|".join(except_list) + ")"

    logger.info(f"Clean OSP simulation case folder: {case_folder}")

    for pattern in case_builder_result_files:
        files = list(case_folder.rglob(pattern))

        for file in files:
            if not re.search(except_pattern, str(file)):
                # logger.info("%s in list to clean" % file)
                if file.is_file():
                    # logger.info("file %s cleaned" % file)
                    file.unlink(missing_ok=True)
                else:
                    # logger.info("dir %s removed" % file)
                    rmtree(file)
    return



from pathlib import Path

from dictIO import DictParser

from ospx import OspCaseBuilder


def test_build():
    # sourcery skip: extract-duplicate-method
    # Prepare
    case_dict_file = Path("test_caseDict_simple")
    parsed_case_dict_file = Path(f"parsed.{case_dict_file.name}")
    _ = DictParser.parse(case_dict_file)
    # Execute
    OspCaseBuilder.build(case_dict_file=parsed_case_dict_file)
    # Assert
    # fmu files
    assert not Path("constantVal.fmu").exists()
    assert not Path("difference.fmu").exists()
    assert not Path("quotient.fmu").exists()
    assert not Path("dividend.fmu").exists()
    assert not Path("subtrahend.fmu").exists()
    assert not Path("minuend.fmu").exists()
    # ModelDescription files-> should NOT have been written
    assert not Path("constantVal_ModelDescription.xml").exists()
    assert not Path("difference_ModelDescription.xml").exists()
    assert not Path("quotient_ModelDescription.xml").exists()
    assert not Path("dividend_ModelDescription.xml").exists()
    assert not Path("subtrahend_ModelDescription.xml").exists()
    assert not Path("minuend_ModelDescription.xml").exists()
    # OspModelDescription files
    assert not Path("constantVal_OspModelDescription.xml").exists()
    assert not Path("difference_OspModelDescription.xml").exists()
    assert not Path("quotient_OspModelDescription.xml").exists()
    assert not Path("dividend_OspModelDescription.xml").exists()
    assert not Path("subtrahend_OspModelDescription.xml").exists()
    assert not Path("minuend_OspModelDescription.xml").exists()
    # SystemStructure files
    assert Path("OspSystemStructure.xml").exists()
    assert Path("SystemStructure.ssd").exists()
    # statisticsDict and watchDict
    assert Path("statisticsDict").exists()
    assert Path("watchDict").exists()


def test_inspect():
    # Prepare
    case_dict_file = Path("test_caseDict")
    parsed_case_dict_file = Path(f"parsed.{case_dict_file.name}")
    _ = DictParser.parse(case_dict_file)
    # Execute
    OspCaseBuilder.build(case_dict_file=parsed_case_dict_file, inspect=True)
    # Assert



from ospx.fmi.experiment import Experiment as Experiment  # noqa: F401
from ospx.fmi.unit import (  # noqa: F401
    Unit as Unit,
    BaseUnit as BaseUnit,
    DisplayUnit as DisplayUnit,
)
from ospx.fmi.variable import (  # noqa: F401
    ScalarVariable as ScalarVariable,
    get_fmi_data_type as get_fmi_data_type,
)
from ospx.fmi.fmu import FMU as FMU  # noqa: F401



import logging
from copy import deepcopy
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, MutableMapping, Union

from dictIO import DictWriter
from dictIO.utils.counter import BorgCounter

from ospx import Connector
from ospx.fmi import FMU, ScalarVariable, Unit

__ALL__ = ["Component"]

logger = logging.getLogger(__name__)


class Component:
    r"""A component is an instance of a (component-) model.

    A component represents an instance of a (component-) model. Any system structure can contain an arbitrary number of components.
    Important here is, that multiple components in a system structure can be instances of one and the same model.
    In practical terms this means that multiple components can refer to the same physical FMU file. \n
    As components are instances of a model (FMU), they inherit the start values defined in the FMU's modelDescription file upon instantiation; \n
    howevere, being an instance, each component can alter and overwrite these start values.
    This is accomplished using the 'initialize' section inside a 'component' element in the ospx case dict. \n
    See https://dnv-opensource.github.io/ospx/fileFormat.caseDict.html

    Equivalent terms to 'component' are: \n
    \t 'Simulator' in OSP. See https://open-simulation-platform.github.io/libcosim/configuration#simulator \n
    \t 'Simulation model' in FMI for co-simulation. See https://github.com/modelica/fmi-standard/releases/download/v2.0.3/FMI-Specification-2.0.3.pdf \n
    \t 'Component' in SSP. See https://ssp-standard.org/publications/SSP10/SystemStructureAndParameterization10.pdf
    """

    def __init__(self, name: str, properties: MutableMapping[Any, Any]):
        self.name: str = name
        self.generate_proxy = False
        self.fmu: FMU
        self.step_size: Union[float, None] = None
        self._initial_values: dict[str, ScalarVariable] = {}
        self._connectors: dict[str, Connector] = {}
        self.generate_proxy: bool = False
        self.remote_access: Union[RemoteAccess, None] = None
        self.counter = BorgCounter()
        self._units: dict[str, Unit]
        self._variables: dict[str, ScalarVariable]

        self._read_fmu(properties)
        self._read_step_size(properties)

        self._read_generate_proxy(properties)
        self._read_remote_access(properties)
        if self.generate_proxy:
            self._generate_proxy()

        self._read_initialize(properties)
        self._init_units()
        self._init_variables()

        self._read_connectors(properties)

    def _read_fmu(self, properties: MutableMapping[Any, Any]):
        if "fmu" not in properties:
            msg = f"component {self.name}: 'fmu' element missing in case dict."
            logger.exception(msg)
            raise ValueError(msg)
        # Read name of fmu file from component properties.
        # Note: Relative paths (if existing) are omitted from the fmu name, as it is by convention expected
        # that the fmu has already been copied from the library into the current working directory (=case folder)
        fmu_file = Path(properties["fmu"])
        if not fmu_file.exists():
            logger.exception(f"component {self.name}: referenced FMU file {fmu_file} not found.")
            raise FileNotFoundError(fmu_file)
        self.fmu = FMU(fmu_file)
        if self.fmu.default_experiment and not self.step_size:
            self.step_size = self.fmu.default_experiment.step_size

    def _read_step_size(self, properties: MutableMapping[Any, Any]):
        if "stepSize" not in properties:
            return
        self.step_size = float(properties["stepSize"])

    def _read_initialize(self, properties: MutableMapping[Any, Any]):
        if "initialize" not in properties:
            return
        for variable_name, variable_properties in properties["initialize"].items():
            variable = ScalarVariable(name=variable_name)
            if "causality" in variable_properties:
                variable.causality = variable_properties["causality"]
            if "variability" in variable_properties:
                variable.variability = variable_properties["variability"]
            if "start" in variable_properties:
                variable.start = variable_properties["start"]
            self._initial_values[variable.name] = variable

    def _read_connectors(self, properties: MutableMapping[Any, Any]):
        if "connectors" not in properties:
            return
        for connector_name, connector_properties in properties["connectors"].items():
            connector = Connector(name=connector_name)
            if "variable" in connector_properties:
                connector.variable = connector_properties["variable"]
            if "variableGroup" in connector_properties:
                connector.variable_group = connector_properties["variableGroup"]
            if "type" in connector_properties:
                connector.type = connector_properties["type"]
            self._connectors[connector.name] = connector

    def _read_generate_proxy(self, properties: MutableMapping[Any, Any]):
        if "generate_proxy" not in properties:
            return
        self.generate_proxy = properties["generate_proxy"]

    def _read_remote_access(self, properties: MutableMapping[Any, Any]):
        if "remoteAccess" not in properties:
            return
        if "host" in properties["remoteAccess"] and "port" in properties["remoteAccess"]:
            self.remote_access = RemoteAccess(
                host=properties["remoteAccess"]["host"],
                port=properties["remoteAccess"]["port"],
            )

    def _generate_proxy(self):
        if not self.remote_access:
            logger.error(
                f"component {self.name}: 'generate_proxy' set to True, but the 'remoteAccess' element is not correctly defined."
            )
        elif not self.remote_access.host:
            logger.error(f"component {self.name}: 'remoteAccess' element is defined, but host is not specified.")
        elif not self.remote_access.port:
            logger.error(f"component {self.name}: 'remoteAccess' element is defined, but port is not specified.")
        else:
            self.fmu = self.fmu.proxify(self.remote_access.host, self.remote_access.port)
            # if NTNU-IHB fmu-proxy code is used, use '-proxy' reference
            self.name = f"{self.name}-proxy"
            # self.name = self.fmu.file.stem

    def _init_units(self):
        self._units = deepcopy(self.fmu.units)

    def _init_variables(self):
        self._variables = deepcopy(self.fmu.variables)

        for variable_name, variable in self._initial_values.items():
            if variable.causality:
                self._variables[variable_name].causality = variable.causality
            if variable.variability:
                self._variables[variable_name].variability = variable.variability
            if variable.start:
                self._variables[variable_name].start = variable.start

    @property
    def variables_with_start_values(self) -> dict[str, ScalarVariable]:
        """Returns a dict with all scalar variables for which start values (initial values)
        are defined in the component.

        Returns
        -------
        dict[str, ScalarVariable]
            dict with all scalar variables with start values defined.
        """
        return self._initial_values

    @property
    def units(self) -> dict[str, Unit]:
        """Returns a dict with all units defined in the component.

        Returns
        -------
        dict[str, Unit]
            dict with all units
        """
        return self._units

    @property
    def variables(self) -> dict[str, ScalarVariable]:
        """Returns a dict with all scalar variables defined in the component.

        Returns
        -------
        dict[str, ScalarVariable]
            dict with all scalar variables
        """
        return self._variables

    @property
    def connectors(self) -> dict[str, Connector]:
        """Returns a dict with all connectors defined by the component.

        Returns
        -------
        dict[str, Connector]
            dict with all connectors
        """
        return self._connectors

    def write_osp_model_description_xml(self):  # sourcery skip: merge-dict-assign
        """Write the <component.name>_OspModelDescription.xml file in the current working directory."""
        osp_model_description_file = self.fmu.file.parent.absolute() / f"{self.name}_OspModelDescription.xml"
        self._clean(osp_model_description_file)

        osp_model_description = {}

        # Unit Definitions
        unit_definitions = {}
        for unit in self.units.values():
            unit_definition = {"_attributes": {}}
            unit_definition["_attributes"]["name"] = unit.name
            if unit.base_unit:
                unit_definition["BaseUnit"] = {"_attributes": {}}
                if unit.base_unit.kg:
                    unit_definition["BaseUnit"]["_attributes"]["kg"] = unit.base_unit.kg
                if unit.base_unit.m:
                    unit_definition["BaseUnit"]["_attributes"]["m"] = unit.base_unit.m
                if unit.base_unit.s:
                    unit_definition["BaseUnit"]["_attributes"]["s"] = unit.base_unit.s
                if unit.base_unit.A:
                    unit_definition["BaseUnit"]["_attributes"]["A"] = unit.base_unit.A
                if unit.base_unit.K:
                    unit_definition["BaseUnit"]["_attributes"]["K"] = unit.base_unit.K
                if unit.base_unit.mol:
                    unit_definition["BaseUnit"]["_attributes"]["mol"] = unit.base_unit.mol
                if unit.base_unit.cd:
                    unit_definition["BaseUnit"]["_attributes"]["cd"] = unit.base_unit.cd
                if unit.base_unit.rad:
                    unit_definition["BaseUnit"]["_attributes"]["rad"] = unit.base_unit.rad
                if unit.base_unit.factor:
                    unit_definition["BaseUnit"]["_attributes"]["factor"] = unit.base_unit.factor
                if unit.base_unit.offset:
                    unit_definition["BaseUnit"]["_attributes"]["offset"] = unit.base_unit.offset
            if unit.display_unit:
                unit_definition["DisplayUnit"] = {"_attributes": {}}
                unit_definition["DisplayUnit"]["_attributes"]["name"] = unit.display_unit.name
                unit_definition["DisplayUnit"]["_attributes"]["factor"] = unit.display_unit.factor
                unit_definition["DisplayUnit"]["_attributes"]["offset"] = unit.display_unit.offset
            unit_definitions[f"{self.counter():06d}_Unit"] = unit_definition
        osp_model_description["UnitDefinitions"] = unit_definitions

        # Variable Groups
        variable_groups = {}
        for variable_name, variable in self.variables.items():
            if not variable.quantity:
                logger.warning(f"component {self.name}: no quantity defined for variable {variable_name}")
            if not variable.unit:
                logger.warning(f"component {self.name}: no unit defined for variable {variable_name}")
            quantity_name = variable.quantity or "UNKNOWN"
            quantity_unit = variable.unit or "UNKNOWN"
            variable_groups[f"{self.counter():06d}_Generic"] = {
                "_attributes": {"name": quantity_name},
                quantity_name: {
                    "_attributes": {"name": quantity_name},
                    "Variable": {
                        "_attributes": {
                            "ref": variable_name,
                            "unit": quantity_unit,
                        }
                    },
                },
            }
        osp_model_description["VariableGroups"] = variable_groups

        # _xmlOpts
        osp_model_description["_xmlOpts"] = {
            "_nameSpaces": {"osp": "https://opensimulationplatform.com/xsd/OspModelDescription-1.0.0.xsd"},
            "_rootTag": "ospModelDescription",
        }

        DictWriter.write(osp_model_description, osp_model_description_file)

    def _clean(self, file_to_remove: Union[str, Path]):
        """Clean up single file."""
        if isinstance(file_to_remove, str):
            file_to_remove = Path.cwd() / file_to_remove
        file_to_remove.unlink(missing_ok=True)


@dataclass()
class RemoteAccess:
    """Data class holding the host and port information for a remote connection."""

    host: str = field(default_factory=lambda: "")
    port: int = 0



# pyright: reportUnnecessaryTypeIgnoreComment=false
import logging
import os
import re
from datetime import datetime as datetime
from pathlib import Path
from typing import Dict, MutableMapping, Union

import matplotlib.pyplot as plt
from matplotlib.figure import Figure

logger = logging.getLogger(__name__)


def create_meta_dict(title: str) -> Dict[str, str]:
    """Create a default of meta dict which can be passed to save_figure().

    Parameters
    ----------
    title : str
        the title of the figure

    Returns
    -------
    Dict[str, str]
        the meta dict
    """
    meta_dict = {
        "Title": title,
        "Author": "VFW",
        "Description": title,
        "Copyright": "VFW",
        "Creation Time": str(datetime.now()),
        "Software": "matplotlib",
        "Disclaimer": "",
        "Warning": "",
        "Source": "",
        "Comment": "",
    }
    return meta_dict


def save_figure(
    fig: Figure,  # type: ignore
    extension: str,
    path: Union[str, os.PathLike[str]],
    title: str,
    meta_dict: MutableMapping[str, str],
):
    """Save a figure object as image file.

    Parameters
    ----------
    fig : Figure
        the Matplotlib figure object
    extension : str
        the file extension. Determines the file format.
    path : Union[str, os.PathLike[str]]
        the folder to save the file in
    title : str
        image title. Will also be used as file name.
    meta_dict : MutableMapping[str, str]
        a dict with additional meta properties. Will be passed as-is to figure.savefig()
    """

    # Make sure path argument is of type Path. If not, cast it to Path type.
    path = path if isinstance(path, Path) else Path(path)

    if not os.path.exists(path):
        logger.info(f"path {path} does not exist, creating")  # 0
        path.mkdir(parents=True, exist_ok=True)

    title_in_file_name = re.sub(r"[\[\]\{\},]+", "", re.sub(r"[\(\)\:\s,]+", "_", title))

    title_string_replacements = [
        ("==", "_eq_"),
        ("!=", "_neq_"),
        (">", "_gt_"),
        ("<", "_lt_"),
        (">=", "_geq_"),
        ("<=", "_leq_"),
        ("+", "_plus_"),
        ("^", "_hat_"),
        ("*", "_ast_"),
        ("|", "_"),
    ]
    for item in title_string_replacements:
        title_in_file_name = title_in_file_name.replace(item[0], item[1])

    # limit overall length to 128 characters
    if len(title_in_file_name) >= 80:
        title_in_file_name = "".join(list(title_in_file_name[:59]) + [".", "."] + list(title_in_file_name[-19:]))

    save_file: str
    if path:
        save_file = str(path / f"{title_in_file_name}.{extension}")
    else:
        save_file = f"{title_in_file_name}.{extension}"

    fig.savefig(  # type: ignore
        save_file,
        orientation="landscape",
        # papertype = 'a4',
        format=extension,
        transparent=False,
        metadata=meta_dict,
    )
    plt.close(fig)  # type: ignore

    return



import logging
import re
from pathlib import Path
from shutil import copy2
from typing import Any, Dict, List, Union

from dictIO import CppDict, DictWriter, XmlFormatter
from dictIO.utils.counter import BorgCounter
from dictIO.utils.path import relative_path

from ospx import Simulation, System
from ospx.utils.dict import find_key

__ALL__ = ["OspSimulationCase"]

logger = logging.getLogger(__name__)


class OspSimulationCase:
    """OSP Simulation Case."""

    def __init__(
        self,
        case_dict: CppDict,
    ):
        self.counter = BorgCounter()
        self.case_dict: CppDict = case_dict
        self.case_folder: Path = case_dict.source_file.resolve().parent if case_dict.source_file else Path.cwd()
        self.system_structure: System

        # Global settings
        self.simulation: Simulation  # general properties of the simulation case
        self._read_simulation()
        self.name: str = self.case_dict.name  # initialize conservatively (with fallback path)
        if self.simulation and self.simulation.name:
            self.name = self.simulation.name

        # Library source path
        self.lib_source: Path
        self._resolve_lib_source_folder()

    def setup(self):
        """Set up the OSP simulation case folder.

        Raises
        ------
        ValueError
            If an expected element in caseDict is missing
        FileNotFoundError
            If an FMU file referenced in caseDict does not exist
        """
        logger.info(f"Set up OSP simulation case '{self.name}' in case folder: {self.case_folder}")

        # Check whether all referenced FMUs actually exist
        self._check_all_fmus_exist()

        # Resolve all referenced FMUs and ensure they are accessible from the case folder via a relative path.
        # This is necessary because OspSystemStructure.xml allows only relative paths
        # as 'source' attribute in a <Simulator> element.
        # If an FMU is not accessible via a relative path, it will be copied into the case folder.
        self._resolve_all_fmus()

        # Read system structure
        if "systemStructure" not in self.case_dict:
            msg = f"no 'systemStructure' section found in {self.case_dict.name}. Cannot set up OSP simulation case."
            logger.exception(msg)
            raise ValueError(msg)
        self.system_structure = System(self.case_dict["systemStructure"])

        # Make sure all components have a step size defined
        self._check_components_step_size()

    def _write_osp_model_description_xmls(self):
        """Write the <component.name>_OspModelDescription.xml files for all components defined in the system structure."""
        logger.info(
            f"Write OspModelDescription.xml files for OSP simulation case '{self.name}' in case folder: {self.case_folder}"
        )
        if not self.system_structure or not self.system_structure.components:
            return
        for component in self.system_structure.components.values():
            component.write_osp_model_description_xml()
        return

    def write_osp_system_structure_xml(self):
        """Write the OspSystemStructure.xml file."""
        # sourcery skip: class-extract-method, merge-dict-assign

        osp_system_structure_file = self.case_folder / "OspSystemStructure.xml"
        self._clean(osp_system_structure_file)

        logger.info(
            f"Write OspSystemStructure.xml file for OSP simulation case '{self.name}' in case folder: {self.case_folder}"
        )

        osp_system_structure: Dict[str, Any] = {}
        osp_system_structure["_xmlOpts"] = {
            "_nameSpaces": {"osp": "https://opensimulationplatform.com/xsd/OspModelDescription-1.0.0.xsd"},
            "_rootTag": "OspSystemStructure",
        }

        # Global Settings
        if self.simulation:
            if self.simulation.start_time:
                osp_system_structure["StartTime"] = self.simulation.start_time
            if self.simulation.base_step_size:
                osp_system_structure["BaseStepSize"] = self.simulation.base_step_size
            if self.simulation.algorithm:
                osp_system_structure["Algorithm"] = self.simulation.algorithm

        # Simulators (=Components)
        simulators: Dict[str, Any] = {}
        for index, (_, component) in enumerate(self.system_structure.components.items()):
            simulator_key = f"{index:06d}_Simulator"
            simulator_properties: Dict[str, Dict[str, Union[str, float, Dict[str, Any], Path]]] = {
                "_attributes": {
                    "name": component.name,
                    "source": relative_path(self.case_folder, component.fmu.file),
                }
            }
            if component.step_size:
                write_step_size_to_osp_system_structure: bool = True
                if (
                    component.fmu.default_experiment
                    and component.fmu.default_experiment.step_size
                    and component.step_size == component.fmu.default_experiment.step_size
                ):
                    write_step_size_to_osp_system_structure = False
                if write_step_size_to_osp_system_structure:
                    simulator_properties["_attributes"]["stepSize"] = component.step_size

            if component.variables_with_start_values:
                simulator_properties["InitialValues"] = {}
                for index, (_, variable) in enumerate(component.variables_with_start_values.items()):
                    if variable.start is not None and variable.data_type is None:
                        logger.error(
                            f"component {component.name}: An initial value is defined for variable {variable.name}, but its data type is not defined.\n"
                            f"The initial value for variable {variable.name} will not be written into OspSystemStructure.xml.\n"
                            "OspSystemStructure.xml will be potentially wrong or incomplete."
                        )
                    else:
                        initial_value_key = f"{index:06d}_InitialValue"
                        initial_value_properties: Dict[str, Any] = {}
                        initial_value_properties["_attributes"] = {"variable": variable.name}
                        if variable.data_type:
                            initial_value_properties[variable.data_type] = {"_attributes": {"value": variable.start}}

                        simulator_properties["InitialValues"][initial_value_key] = initial_value_properties
            simulators[simulator_key] = simulator_properties

        osp_system_structure["Simulators"] = simulators

        # Connections
        connections: Dict[str, Dict[str, Any]] = {}
        for connection in self.system_structure.connections.values():
            if not connection.is_valid:
                continue
            if connection.is_variable_connection:
                connection_key = f"{self.counter():06d}_VariableConnection"
                # (note: the order 000000, 000001 is essential here!)
                connections[connection_key] = {
                    "000000_Variable": {
                        "_attributes": {
                            "simulator": connection.source_endpoint.component.name,
                            "name": connection.source_endpoint.variable_name,
                        }
                    },
                    "000001_Variable": {
                        "_attributes": {
                            "simulator": connection.target_endpoint.component.name,
                            "name": connection.target_endpoint.variable_name,
                        }
                    },
                }
            if connection.is_variable_group_connection:
                connection_key = f"{self.counter():06d}_VariableGroupConnection"
                # (note: the order 000000, 000001 is essential here!)
                connections[connection_key] = {
                    "000000_VariableGroup": {
                        "_attributes": {
                            "simulator": connection.source_endpoint.component.name,
                            "name": connection.source_endpoint.variable_name,
                        }
                    },
                    "000001_VariableGroup": {
                        "_attributes": {
                            "simulator": connection.target_endpoint.component.name,
                            "name": connection.target_endpoint.variable_name,
                        }
                    },
                }
        osp_system_structure["Connections"] = connections

        # Write OspSystemStructure.xml
        formatter = XmlFormatter()
        DictWriter.write(osp_system_structure, osp_system_structure_file, formatter=formatter)

        self._correct_wrong_xml_namespace(
            "OspSystemStructure.xml",
            "<OspSystemStructure.*>?",
            """<OspSystemStructure xmlns="http://opensimulationplatform.com/MSMI/OSPSystemStructure" version="0.1">""",
        )

        return

    def write_system_structure_ssd(self):
        """Write the SystemStructure.ssd file."""

        system_structure_ssd_file = self.case_folder / "SystemStructure.ssd"
        self._clean(system_structure_ssd_file)

        # sourcery skip: merge-dict-assign
        logger.info(
            f"Write SystemStructure.ssd file for OSP simulation case '{self.name}' in case folder: {self.case_folder}"
        )

        system_structure_ssd: Dict[str, Any] = {}
        system_structure_ssd["_xmlOpts"] = {
            "_nameSpaces": {
                "ssd": "file:///C:/Software/OSP/xsd/SystemStructureDescription",
                "ssv": "file:///C:/Software/OSP/xsd/SystemStructureParameterValues",
                "ssc": "file:///C:/Software/OSP/xsd/SystemStructureCommon",
            },
            "_rootTag": "SystemStructureDescription",
        }
        system_structure_ssd["System"] = {
            "_attributes": {
                "name": self.name,
                "description": self.name,
            }
        }

        # Global settings
        default_experiment = {
            "Annotations": {
                "Annotation": {
                    "_attributes": {"type": "com.opensimulationplatform"},
                    "Algorithm": {
                        "FixedStepAlgorithm": {
                            "_attributes": {
                                "baseStepSize": str(self.simulation.base_step_size),
                                "startTime": str(self.simulation.start_time),
                                "stopTime": str(self.simulation.stop_time),
                            }
                        }
                    },
                }
            }
        }
        system_structure_ssd["DefaultExperiment"] = default_experiment

        # Components
        components: Dict[str, Any] = {}
        for component_name, component in self.system_structure.components.items():
            connectors: Dict[str, Dict[str, Any]] = {}
            for connector in component.connectors.values():
                if connector.variable and connector.type:
                    connector_key = f"{self.counter():06d}_Connector"
                    # (note: the order 000000, 000001 is essential here!)
                    connectors[connector_key] = {
                        "_attributes": {
                            "name": connector.variable,
                            "kind": connector.type,
                        },
                        "Real": {},
                    }
            element_key = f"{self.counter():06d}_Component"
            components[element_key] = {
                "_attributes": {
                    "name": component_name,
                    "source": relative_path(self.case_folder, component.fmu.file),
                },
                "Connectors": connectors,
            }
        system_structure_ssd["System"]["Elements"] = components

        # Connections
        connections: Dict[str, Any] = {}
        for connection in self.system_structure.connections.values():
            if connection.source_endpoint and connection.target_endpoint:
                connection_key = f"{self.counter():06d}_Connection"
                connections[connection_key] = {
                    "_attributes": {
                        "startElement": connection.source_endpoint.component.name,
                        "startConnector": connection.source_endpoint.variable_name,
                        "endElement": connection.target_endpoint.component.name,
                        "endConnector": connection.target_endpoint.variable_name,
                    }
                }
        system_structure_ssd["System"]["Connections"] = connections

        # Write SystemStructure.ssd
        formatter = XmlFormatter(omit_prefix=False)
        DictWriter.write(system_structure_ssd, system_structure_ssd_file, formatter=formatter)

        return

    def write_statistics_dict(self):
        """Write selected properties of the system structure into a statistics dict.

        I.e. for documentation or further statistical analysis.
        """
        statistics_dict_file = self.case_folder / "statisticsDict"
        # self._clean(statistics_dict_file)

        # sourcery skip: merge-dict-assign, simplify-dictionary-update
        logger.info(f"Write statistics dict for OSP simulation case '{self.name}' in case folder: {self.case_folder}")

        statistics_dict = {}

        statistics_dict["simulation"] = {"name": self.simulation.name}

        statistics_dict["components"] = {
            "count": len(self.system_structure.components.keys()),
            "names": list(self.system_structure.components.keys()),
        }

        statistics_dict["connections"] = {
            "count": len(self.system_structure.connections.keys()),
            "names": list(self.system_structure.connections.keys()),
        }

        statistics_dict["connectors"] = {
            "count": len(self.system_structure.connectors.keys()),
            "names": list(self.system_structure.connectors.keys()),
        }

        unit_list: List[str] = []
        display_unit_list: List[str] = []
        factors_list: List[float] = []
        offsets_list: List[float] = []
        for unit in self.system_structure.units.values():
            unit_list.append(unit.name)
            display_unit_list.append(unit.display_unit.name)
            factors_list.append(unit.display_unit.factor)
            offsets_list.append(unit.display_unit.offset)

        statistics_dict["units"] = {
            "count": len(self.system_structure.units.keys()),
            "unitNames": unit_list,
            "displayUnitNames": display_unit_list,
            "factors": factors_list,
            "offsets": offsets_list,
        }

        statistics_dict["variables"] = {
            "count": len(self.system_structure.variables.keys()),
            "names": list(self.system_structure.variables.keys()),
        }

        DictWriter.write(statistics_dict, statistics_dict_file, mode="w")

    def write_watch_dict(self):
        """Write a case-specific watch dict file.

        The watch dict file can be used with watchCosim for
            - convergence control
            - convergence plotting
            - extracting the results
        """
        watch_dict_file = self.case_folder / "watchDict"
        # self._clean(watch_dict_file)

        logger.info(f"Write watch dict for OSP simulation case '{self.name}' in case folder: {self.case_folder}")

        watch_dict: Dict[str, Any] = {
            "datasources": {},
            "delimiter": ",",  # 'objects': {},
            "simulation": {"name": self.simulation.name},
        }

        # @TODO: Time, StepCount, conn0, conn1, etc from modelDescription.xml ModelVariables
        #        should match connectors in caseDict for respective model. Improvement needed.
        #        FRALUM, 2021-xx-xx
        time_column = 0
        # Components
        for component_name, component in self.system_structure.components.items():
            no_of_connectors = len(component.connectors.keys())

            data_columns = [1] + [x + 2 for x in range(no_of_connectors)]  # f*** StepCount
            watch_dict["datasources"].update({component_name: {"dataColumns": data_columns, "timeColumn": time_column}})

        DictWriter.write(watch_dict, watch_dict_file, mode="w")

        return

    def _read_simulation(self):
        """Read general simulation properties from case dict."""
        logger.info("reading simulation properties")  # 0

        if "run" not in self.case_dict:
            return
        if "simulation" not in self.case_dict["run"]:
            return
        simulation = Simulation()
        simulation_properties = self.case_dict["run"]["simulation"]
        if "name" in simulation_properties:
            simulation.name = simulation_properties["name"]
        if "startTime" in simulation_properties:
            simulation.start_time = simulation_properties["startTime"]
        if "stopTime" in simulation_properties:
            simulation.stop_time = simulation_properties["stopTime"]
        if "baseStepSize" in simulation_properties:
            simulation.base_step_size = simulation_properties["baseStepSize"]
        if "algorithm" in simulation_properties:
            simulation.algorithm = simulation_properties["algorithm"]
        self.simulation = simulation

    def _resolve_lib_source_folder(self):
        """Resolve the library source folder."""
        self.lib_source = Path.cwd()  # initialize conservatively (with fallback path)
        if "_environment" in self.case_dict:
            if "libSource" in self.case_dict["_environment"]:
                self.lib_source = Path(self.case_dict["_environment"]["libSource"])
            else:
                logger.warning(
                    f"no 'libSource' element found in {self.case_dict.name}['_environment']. Path to libSource will be set to current working directory."
                )
        else:
            logger.warning(
                f"no '_environment' section found in {self.case_dict.name}. Path to libSource hence is unknown and will be set to current working directory."
            )
        self.lib_source = self.lib_source.resolve().absolute()

    def _resolve_fmu_file(self, fmu_name: str) -> Path:
        fmu_file: Path = Path(fmu_name)
        if fmu_file.is_absolute():
            fmu_file = fmu_file.resolve()
        else:
            fmu_file = (self.lib_source / fmu_file).resolve()
        return fmu_file

    def _check_all_fmus_exist(self):
        """Check whether all referenced FMUs actually exist."""
        logger.debug("Check whether all referenced FMUs exist.")
        components = self.case_dict["systemStructure"]["components"]

        for component_name, component_properties in components.items():
            if "fmu" not in component_properties:
                msg = f"component {component_name}: 'fmu' element missing in case dict."
                logger.exception(msg)
                raise ValueError(msg)
            fmu_file = self._resolve_fmu_file(component_properties["fmu"])
            if not fmu_file.exists():
                msg = f"component {component_name}: referenced FMU file {fmu_file} not found."
                logger.exception(msg)
                raise FileNotFoundError(fmu_file)

    def _resolve_all_fmus(self):
        """Resolve all referenced FMUs and ensures they are accessible from the case folder via a relative path.

        This is necessary because OspSystemStructure.xml allows only relative paths
        as 'source' attribute in a <Simulator> element.
        If an FMU is not accessible via a relative path, the FMU will be copied into the case folder.
        Note: If multiple components reference the same FMU, these get copied only once.
        """

        logger.debug("Ensure all referenced FMUs are accessible from the case folder via a relative path.")
        components = self.case_dict["systemStructure"]["components"]
        for _, component_properties in components.items():
            fmu_file = self._resolve_fmu_file(component_properties["fmu"])
            try:
                _ = relative_path(self.case_folder, fmu_file)
            except ValueError:
                fmu_file = self._copy_fmu_to_case_folder(fmu_file)
            component_properties["fmu"] = fmu_file

    def _copy_fmu_to_case_folder(self, fmu_file: Path) -> Path:
        """Copy the passed in FMU file into the case folder.

        If also an accompanying <fmu_name>_OspModelDescription.xml file exists in the same folder as the FMU file,
        then also that OspModelDescription.xml file will be copied into the case folder.

        Parameters
        ----------
        fmu_file : Path
            FMU file to be copied into the case folder.

        Returns
        -------
        Path
            FMU file copied into the case folder.
        """
        fmu_file_in_case_folder: Path = (self.case_folder / fmu_file.name).resolve().absolute()
        if not fmu_file_in_case_folder.exists():
            logger.info(f"Copy FMU {fmu_file} --> {fmu_file_in_case_folder}")
            copy2(fmu_file, self.case_folder)
            # Check whether also an <fmu_name>_OspModelDescription.xml file exists.
            # If so, copy also that one.
            osp_model_description_file = fmu_file.with_name(f"{fmu_file.stem}_OspModelDescription.xml")
            if osp_model_description_file.exists():
                logger.info(f"Copy OspModelDescription {osp_model_description_file} --> {fmu_file_in_case_folder}")
                copy2(osp_model_description_file, self.case_folder)
        return fmu_file_in_case_folder

    def _check_components_step_size(self):
        """Ensure that all components have a step size defined.

        If a components step size is undefined, it will be set to the base step size.
        """
        if not self.system_structure or not self.system_structure.components:
            return
        if not self.simulation or not self.simulation.base_step_size:
            return
        for component in self.system_structure.components.values():
            if not component.step_size:
                component.step_size = self.simulation.base_step_size
        return

    def _set_components_step_size(self, step_size: float):
        """Overwrite the step size of all components with the passed in value."""
        if not self.system_structure or not self.system_structure.components:
            return
        for component in self.system_structure.components.values():
            component.step_size = step_size
        return

    def _inspect(self):
        """Inspects all components and all FMUs for the public variable names and units they declare, as documented in their modelDescription.xml's.

        Results get logged to the console.
        """
        logger.info(f"Inspect OSP simulation case '{self.name}' in case folder: {self.case_folder}")

        delim = "\t" * 3

        log_string = (
            f"Components and related FMUs as defined in {self.case_dict.name}\n" f"\tcomponent{delim}fmu{delim}\n\n"
        )
        for component_name, component in self.system_structure.components.items():
            log_string += f"\t{component_name}{delim}{component.fmu.file.name}\n"
        logger.info(log_string + "\n")

        log_string = f"FMU attributes defined in the fmu's modelDescription.xml\n" f"\tfmu{delim}attributes{delim}"
        for fmu_name, fmu in self.system_structure.fmus.items():
            log_string += f"\n\n\t{fmu_name}\n"
            fmu_attributes = "\n".join(
                f"\t{delim}{k}{delim}{v}" for k, v in fmu.model_description["_xmlOpts"]["_rootAttributes"].items()
            )
            log_string += fmu_attributes
            if default_experiment_key := find_key(fmu.model_description, "DefaultExperiment$"):
                if "_attributes" in fmu.model_description[default_experiment_key]:
                    fmu_default_experiment = "\n".join(
                        f"\t{delim}{k}{delim}{v}"
                        for k, v in fmu.model_description[default_experiment_key]["_attributes"].items()
                    )
                    log_string += f"\n{fmu_default_experiment}"
        logger.info(log_string + "\n")

        log_string = (
            f"Unit definitions defined in the fmu's modelDescription.xml\n"
            f"\tfmu{delim}unit{delim}display unit{delim}factor{delim}offset"
        )
        for fmu_name, fmu in self.system_structure.fmus.items():
            log_string += f"\n\n\t{fmu_name}\n"
            unit_definitions = "\n".join(
                f"\t{delim}{unit_name}{delim}{unit.display_unit.name}\t{delim}{unit.display_unit.factor}{delim}{unit.display_unit.offset}"
                for unit_name, unit in fmu.units.items()
            )
            log_string += unit_definitions
        logger.info(log_string + "\n")

        log_string = (
            f"Variables defined in the fmu's modelDescription.xml\n" f"\tfmu{delim}variable{delim}type{delim}unit"
        )
        logger.info(log_string + "\n")
        for fmu_name, fmu in self.system_structure.fmus.items():
            log_string = f"\t{fmu_name}\n"
            variable_definitions = "\n".join(
                f"\n\n\t{delim}{variable_name}{delim}{variable.data_type}{delim}{variable.unit}"
                for variable_name, variable in fmu.variables.items()
            )
            log_string += variable_definitions
            logger.info(log_string + "\n")

        log_string = (
            f"Connectors defined in {self.case_dict.name}\n"
            f"\tComponent{delim}Connector{delim}Variable{delim}VariableGroup{delim}Type"
        )
        for component_name, component in self.system_structure.components.items():
            if component.connectors:
                log_string += f"\n\n\t{component_name}\n"
                connector_definitions = "\n".join(
                    f"\t{delim}{connector_name}{delim}{connector.variable}{delim}{connector.variable_group}{delim}{connector.type}"
                    for connector_name, connector in component.connectors.items()
                )
                log_string += connector_definitions
        logger.info(log_string + "\n")

        logger.info("Inspect mode: Finished.")

    def _write_plot_config_json(self):
        """Write the PlotConfig.json file, containing postprocessing information."""

        plot_config_file = self.case_folder / "PlotConfig.json"
        self._clean(plot_config_file)

        if "plots" in self.case_dict["postProcessing"].keys():
            temp_dict: Dict[str, List[Dict[str, Any]]] = {"plots": []}
            for plot in self.case_dict["postproc"]["plots"].values():
                variables: List[Dict[str, Any]] = []
                for (
                    component_name,
                    component,
                ) in self.system_structure.components.items():
                    variables.extend(
                        {
                            "simulator": component_name,
                            "variable": connector.variable,
                        }
                        for connector_name, connector in component.connectors.items()
                        if connector_name in plot["ySignals"]
                    )

                temp_dict["plots"].append(
                    {
                        "label": plot["title"],
                        "plotType": "trend",
                        "variables": variables,
                    }
                )

            DictWriter.write(temp_dict, plot_config_file)

        return

    def _correct_wrong_xml_namespace(self, file_name: str, pattern: str, replace: str):
        """Substitutes namespace
        (may be obsolete in future).
        """
        buffer = ""
        with open(file_name, "r") as f:
            buffer = re.sub(pattern, replace, f.read())

        with open(file_name, "w") as f:
            _ = f.write(buffer)

        return

    def _clean(self, file_to_remove: Union[str, Path]):
        """Clean up single file."""
        if isinstance(file_to_remove, str):
            file_to_remove = self.case_folder / file_to_remove
        file_to_remove.unlink(missing_ok=True)



from ospx.simulation import Simulation as Simulation  # noqa: F401
from ospx.connector import Connector as Connector  # noqa: F401
from ospx.connection import (  # noqa: F401
    Endpoint as Endpoint,
    Connection as Connection,
)
from ospx.component import Component as Component  # noqa: F401
from ospx.system import System as System  # noqa: F401
from ospx.ospSimulationCase import OspSimulationCase as OspSimulationCase  # noqa: F401
from ospx.graph import Graph as Graph  # noqa: F401
from ospx.ospCaseBuilder import OspCaseBuilder as OspCaseBuilder  # noqa: F401
from ospx.importer import OspSystemStructureImporter as OspSystemStructureImporter  # noqa: F401



# pyright: reportUnknownMemberType=false
# pyright: reportUnknownParameterType=false
# pyright: reportUnnecessaryTypeIgnoreComment=false

import functools
import logging
import re
from typing import Any, Dict, Tuple, Union

from graphviz import Digraph
from graphviz.dot import Dot

from ospx import Component, Connection, OspSimulationCase

__ALL__ = ["Graph"]

logger = logging.getLogger(__name__)


class Graph:
    """Class providing methods to generate a visual dependency graph
    of a system's components and its connections.
    """

    @staticmethod
    def generate_dependency_graph(case: OspSimulationCase):
        """Generate a dependency graph of the system structure as pdf, for documentation.

        Note: This requires graphviz to be installed on the local machine
        """
        graphiz_not_found_error_mesage: str = (
            "OspSimulationCase.generate_dependency_graph(): failed to run graphviz. \n"
            "To generate the system structure dependency graph, graphviz needs to be installed on the local machine. \n"
            "Kindly check your local installation of graphviz."
        )

        # Default styles
        text_size: str = "11"
        styles: Dict[str, Dict[str, str]] = {
            "graph": {
                "label": f"{case.simulation.name}",
                "fontname": "Verdana",
                "fontsize": text_size,
                "fontcolor": "black",
                "bgcolor": "white",
                "rankdir": "TD",
                "overlap": "compress",
                "sep": "10,100",
                "remincross": "true",
                "ratio": "fill",
                "margin": "0",
                "size": "10, 10!",
            },
            "nodes": {
                "fontname": "Verdana",
                "fontsize": text_size,
                "fontcolor": "white",
                "shape": "square",
                "color": "magenta",
                "style": "filled",
                "fillcolor": "magenta",
            },
            "edges": {
                "style": "dashed",
                "color": "magenta",
                "penwidth": "3",
                "arrowhead": "open",
                "fontname": "Verdana",
                "fontsize": text_size,
                "fontcolor": "magenta",
            },
        }

        basic_op_names: str = "(power|dot|sum|diff|prod|div|quotient)"
        input_names: str = "^(INP|inp)"

        callgraph: Dot
        try:
            digraph = functools.partial(Digraph, format="png")
            callgraph = digraph()
            callgraph = _apply_styles(callgraph, styles)
        except Exception:
            logger.exception(graphiz_not_found_error_mesage)
            return

        # Components
        for component in case.system_structure.components.values():
            label_key, label = _get_node_label(component)
            # var_keys = find_key(case.models[key]['InitialValues'], 'InitialValue')
            # variables = {}
            label = _create_table(
                label_key,
                {
                    "source:": component.fmu.file.name,
                    "stepsize:": component.step_size,
                    "variables:": "",
                },
            )

            if re.search(input_names, component.name):
                shape = "diamond"
                style = "filled,rounded"
                fillcolor = "#FFFFFF"
            elif re.search(basic_op_names, component.name, re.I):
                label = label
                shape = "square"
                style = "filled, rounded"
                fillcolor = "#EEBBDD"
            else:
                shape = "square"
                style = "filled"
                fillcolor = "#DDDDEE"

            callgraph.node(
                label_key,
                label=label,
                fontname="Verdana",
                fontsize=text_size,
                fontcolor="black",
                shape=shape,
                color="black",
                style=style,
                fillcolor=fillcolor,
            )

        # Connections

        for _, connection in case.system_structure.connections.items():
            if not (connection.source_endpoint and connection.target_endpoint):
                return
            if not (connection.source_endpoint.component and connection.target_endpoint.component):
                return
            from_key: str = connection.source_endpoint.component.name
            to_key: str = connection.target_endpoint.component.name

            label = _get_edge_label(connection)

            if re.search(input_names, from_key, re.I):
                label = "input\n%s" % label
                style = "dashed"
                color = "#003399"
                fontcolor = "#003399"
                penwidth = ("%i" % 1,)
                weight = "%i" % 1

            elif re.search(basic_op_names, from_key, re.I):
                style = "filled"
                color = "#995566"
                fontcolor = "#663344"
                penwidth = ("%i" % 3,)
                weight = ("%.2f" % 0.66,)

            else:
                style = "bold"
                color = "black"
                fontcolor = "black"
                penwidth = ("%i" % int(round((2) ** 1.5, 0)),)
                weight = ("%i" % int(round((2) ** 1.5, 0)),)

            callgraph.edge(
                from_key,
                to_key,
                style=style,
                color=color,
                arrowhead="open",
                fontname="Verdana",
                fontsize=text_size,
                fontcolor=fontcolor,
                penwidth=str(penwidth),
                weight=str(weight),
                label=label,
                overlap="false",
                splines="true",
            )

        # Create callGraph pdf

        try:
            callgraph.render(f"{case.simulation.name}_callGraph", format="pdf")  # type: ignore
        except Exception:
            logger.exception(graphiz_not_found_error_mesage)

        return


def _apply_styles(digraph: Dot, styles: Dict[str, Any]) -> Dot:
    digraph.graph_attr.update(("graph" in styles and styles["graph"]) or {})  # type: ignore
    digraph.node_attr.update(("nodes" in styles and styles["nodes"]) or {})  # type: ignore
    digraph.edge_attr.update(("edges" in styles and styles["edges"]) or {})  # type: ignore
    return digraph


def _get_node_label(component: Component) -> Tuple[str, str]:
    label = f"{component.name}\n___________\n\nfmu\n"
    label += re.sub(r"(^.*/|^.*\\|\.fmu.*$)", "", component.fmu.file.name)

    label_key = component.name
    return label_key, label


def _get_edge_label(connection: Connection) -> str:
    return (
        f"{connection.source_endpoint.variable_name}-->{connection.target_endpoint.variable_name}"
        if connection.is_valid
        else ""
    )


def _create_table(name: str, child: Union[Dict[str, Any], None] = None) -> str:
    _child: Dict[str, Any] = child or {" ": " "}
    n_child = len(_child)
    string: str = (
        f'<\n<TABLE BORDER="1" CELLBORDER="1" CELLSPACING="0">\n<TR>\n<TD COLSPAN="{2 * n_child:d}">{name}</TD>\n</TR>\n'
    )
    for key, item in _child.items():
        string += f"<TR><TD>{key}</TD><TD>{item}</TD></TR>\n"
    string += "</TABLE>\n>"

    return string



import logging
import os
import platform
import re
from copy import deepcopy
from datetime import date, datetime
from pathlib import Path
from shutil import copyfile
from typing import Any, Dict, Mapping, MutableMapping, Union
from zipfile import ZipFile

from dictIO import CppDict, XmlFormatter, XmlParser
from dictIO.utils.counter import BorgCounter

from ospx.fmi import BaseUnit, DisplayUnit, Experiment, ScalarVariable, Unit
from ospx.utils.dict import find_key, find_type_identifier_in_keys, shrink_dict
from ospx.utils.zip import (
    add_file_content_to_zip,
    read_file_content_from_zip,
    remove_files_from_zip,
    rename_file_in_zip,
)

__ALL__ = ["FMU"]

logger = logging.getLogger(__name__)


class FMU:
    """Class to read and interact with an fmi 2.0 Functional Mockup Unit (FMU).

    See also https://github.com/modelica/fmi-standard/blob/v2.0.x/schema/fmi2ModelDescription.xsd
    """

    def __init__(self, file: Union[str, os.PathLike[str]]):
        # Make sure fmu_file argument is of type Path. If not, cast it to Path type.
        file = file if isinstance(file, Path) else Path(file)
        if not file.exists():
            logger.error(f"DictParser: File {file} not found.")
            raise FileNotFoundError(file)

        self.file: Path = file
        self.model_description: CppDict = self._read_model_description()
        self.counter = BorgCounter()

    def _read_model_description(self) -> CppDict:
        model_description = CppDict(Path("modelDescription.xml"))
        xml_parser = XmlParser()

        logger.info(f"{self.file.name}: read modelDescription.xml")

        if file_content := read_file_content_from_zip(self.file, "modelDescription.xml"):
            model_description = xml_parser.parse_string(file_content, model_description)

        self._clean_solver_internal_variables(model_description)

        self.model_description = model_description

        return model_description

    def _write_model_description(
        self,
        model_description: Union[CppDict, None] = None,
        write_inside_fmu: bool = False,
    ):
        """Save updated model_description both inside FMU as well as separate file in the FMUs directory."""
        if model_description:
            self.model_description = model_description

        self.model_description["_xmlOpts"]["_nameSpaces"] = {
            "xs": "file:///C:/Software/OSP/xsd/fmi3ModelDescription.xsd"
        }

        formatter = XmlFormatter()
        formatted_xml = formatter.to_string(self.model_description)

        logger.info(f"{self.file.name}: write modelDescription.xml")

        # Write internal modelDescription.xml (inside FMU)
        if write_inside_fmu:
            _ = remove_files_from_zip(self.file, "modelDescription.xml")
            _ = add_file_content_to_zip(self.file, "modelDescription.xml", formatted_xml)

        # Write external modelDescription.xml (separate file, beside FMU)
        external_file = self.file.parent.absolute() / f"{self.file.stem}_ModelDescription.xml"
        with open(external_file, "w") as f:
            _ = f.write(formatted_xml)

        return

    @property
    def units(self) -> Dict[str, Unit]:
        """Returns a dict with all units defined in the FMU.

        Returns
        -------
        Dict[str, Unit]
            dict with all units
        """
        model_unit_definitions: MutableMapping[Any, Any] = {}
        if unit_definitions_key := find_key(self.model_description, "UnitDefinitions$"):
            model_unit_definitions = self.model_description[unit_definitions_key]
            # make sure unit definitions are unique (e.g. to keep XML files clean)
            model_unit_definitions = shrink_dict(model_unit_definitions, unique_key=["_attributes", "name"])
        unit_definitions: dict[str, Unit] = {}
        for u in model_unit_definitions.values():
            unit = Unit()
            unit.name = u["_attributes"]["name"]
            # BaseUnit
            if base_unit_key := find_key(u, "BaseUnit$"):
                unit.base_unit = BaseUnit()
                if "kg" in u[base_unit_key]["_attributes"]:
                    unit.base_unit.kg = u[base_unit_key]["_attributes"]["kg"]
                if "m" in u[base_unit_key]["_attributes"]:
                    unit.base_unit.m = u[base_unit_key]["_attributes"]["m"]
                if "s" in u[base_unit_key]["_attributes"]:
                    unit.base_unit.s = u[base_unit_key]["_attributes"]["s"]
                if "A" in u[base_unit_key]["_attributes"]:
                    unit.base_unit.A = u[base_unit_key]["_attributes"]["A"]
                if "K" in u[base_unit_key]["_attributes"]:
                    unit.base_unit.K = u[base_unit_key]["_attributes"]["K"]
                if "mol" in u[base_unit_key]["_attributes"]:
                    unit.base_unit.mol = u[base_unit_key]["_attributes"]["mol"]
                if "cd" in u[base_unit_key]["_attributes"]:
                    unit.base_unit.cd = u[base_unit_key]["_attributes"]["cd"]
                if "rad" in u[base_unit_key]["_attributes"]:
                    unit.base_unit.rad = u[base_unit_key]["_attributes"]["rad"]
                if "factor" in u[base_unit_key]["_attributes"]:
                    unit.base_unit.factor = u[base_unit_key]["_attributes"]["factor"]
                if "offset" in u[base_unit_key]["_attributes"]:
                    unit.base_unit.offset = u[base_unit_key]["_attributes"]["offset"]
            # DisplayUnit
            if display_unit_key := find_key(u, "DisplayUnit$"):
                unit.display_unit = DisplayUnit()
                if "name" in u[display_unit_key]["_attributes"]:
                    unit.display_unit.name = u[display_unit_key]["_attributes"]["name"]
                if "factor" in u[display_unit_key]["_attributes"]:
                    unit.display_unit.factor = u[display_unit_key]["_attributes"]["factor"]
                if "offset" in u[display_unit_key]["_attributes"]:
                    unit.display_unit.offset = u[display_unit_key]["_attributes"]["offset"]
            unit_definitions[unit.name] = unit
        return unit_definitions

    @property
    def variables(self) -> dict[str, ScalarVariable]:
        """Returns a dict with all scalar variables defined in the FMU.

        Returns
        -------
        dict[str, ScalarVariable]
            dict with all scalar variables
        """
        model_variables_key = find_key(self.model_description, "ModelVariables$")
        if not model_variables_key:
            return {}
        # Read model variables from model description
        # (without "_" and "settings" as these are proprietary variables from iti)
        model_variables = {
            k: v
            for k, v in self.model_description[model_variables_key].items()
            if not re.match("^(_|settings)", v["_attributes"]["name"])
        }
        # Translate variable attributes from model description into Variable objects
        variables: dict[str, ScalarVariable] = {}
        for k, v in model_variables.items():
            variable_type: str = re.sub(r"^\d{6}_", "", k)
            if variable_type == "ScalarVariable":
                variable = ScalarVariable(name=v["_attributes"]["name"])
                if "valueReference" in v["_attributes"]:
                    variable.value_reference = v["_attributes"]["valueReference"]
                if "description" in v["_attributes"]:
                    variable.description = v["_attributes"]["description"]
                if "causality" in v["_attributes"]:
                    variable.causality = v["_attributes"]["causality"]
                if "variability" in v["_attributes"]:
                    variable.variability = v["_attributes"]["variability"]
                if type_identifier := find_type_identifier_in_keys(v):
                    variable.data_type = type_identifier
                    type_key = find_key(v, f"{type_identifier}$")
                    if v[type_key] and "_attributes" in v[type_key]:
                        if "quantity" in v[type_key]["_attributes"]:
                            variable.quantity = v[type_key]["_attributes"]["quantity"]
                        if "unit" in v[type_key]["_attributes"]:
                            variable.unit = v[type_key]["_attributes"]["unit"]
                        if "display_unit" in v[type_key]["_attributes"]:
                            variable.display_unit = v[type_key]["_attributes"]["display_unit"]
                        if "start" in v[type_key]["_attributes"]:
                            variable.start = v[type_key]["_attributes"]["start"]
                variables[variable.name] = variable

        return variables

    @property
    def default_experiment(self) -> Union[Experiment, None]:
        """Returns the default experiment, if defined in the FMU.

        Returns
        -------
        Union[Experiment, None]
            the default experiment, if defined. Otherwise None.
        """
        default_experiment_key = find_key(self.model_description, "DefaultExperiment$")
        if not default_experiment_key:
            return None
        default_experiment = Experiment()
        default_experiment_properties = self.model_description[default_experiment_key]
        if "_attributes" in default_experiment_properties:
            default_experiment_attributes = default_experiment_properties["_attributes"]
            if "startTime" in default_experiment_attributes:
                default_experiment.start_time = default_experiment_attributes["startTime"]
            if "stopTime" in default_experiment_attributes:
                default_experiment.stop_time = default_experiment_attributes["stopTime"]
            if "tolerance" in default_experiment_attributes:
                default_experiment.tolerance = default_experiment_attributes["tolerance"]
            if "stepSize" in default_experiment_attributes:
                default_experiment.step_size = default_experiment_attributes["stepSize"]
        return default_experiment

    def copy(self, new_name: str):
        """Save a copy of the FMU with a new name.

        Parameters
        ----------
        new_name : str
            Intended name of the copy. The new name must be different from the existing name.

        Returns
        -------
        FMU
            The new FMU
        """
        # Prepare
        new_name = Path(new_name).stem
        existing_file_name = self.file.stem
        if new_name == existing_file_name:
            logger.error(f"{self.file.name} copy: new name {new_name} is identical with existing name. copy() aborted.")
        new_model_description: CppDict = deepcopy(self.model_description)
        new_file = self.file.parent.absolute() / f"{new_name}.fmu"

        # Copy FMU
        _ = copyfile(self.file, new_file)

        # Rename *.dll files in FMU to match new fmu name
        with ZipFile(new_file, "r") as document:
            dll_file_names = [
                file.filename
                for file in document.infolist()
                if re.search(r".*\.dll$", file.filename) and existing_file_name in file.filename
            ]
        new_dll_file_names = [re.sub(existing_file_name, new_name, dll_file_name) for dll_file_name in dll_file_names]
        for dll_file_name, new_dll_file_name in zip(dll_file_names, new_dll_file_names):
            logger.info(f"{self.file.name} copy: renaming dll {dll_file_name} to {new_dll_file_name}")
            _ = rename_file_in_zip(new_file, dll_file_name, new_dll_file_name)

        # Rename <fmiModelDescription modelName> in modelDescription.xml
        new_model_description["_xmlOpts"]["_rootAttributes"]["modelName"] = new_name

        # Rename <CoSimulation modelIdentifier> in modelDescription.xml
        # (STC requires consistency between <fmiModelDescription modelName> and <CoSimulation modelIdentifier>)
        co_simulation: MutableMapping[Any, Any] = new_model_description[
            find_key(new_model_description, "CoSimulation$")
        ]
        co_simulation["_attributes"]["modelIdentifier"] = new_name

        # Log the update in modelDescription.xml
        self._log_update_in_model_description(new_model_description)

        # Write updated modelDescription.xml into new FMU
        new_fmu = FMU(new_file)
        new_fmu._write_model_description(new_model_description)

        return new_fmu

    def proxify(self, host: str, port: int):
        """Create a proxy version of the FMU.

        For details see https://github.com/NTNU-IHB/FMU-proxy

        Parameters
        ----------
        host : str
            Remote host
        port : int
            Remote port

        Returns
        -------
        FMU
            The created proxy version of the FMU
        """

        import subprocess

        remote_string = f"--remote={host}:{port}"
        command = f"fmu-proxify {self.file.name} {remote_string}"
        try:
            _ = subprocess.run(command, timeout=60)
        except subprocess.TimeoutExpired:
            logger.exception(f"Timeout occured when calling {command}.")
            return self
        proxy_fmu_file = self.file.parent.absolute() / f"{self.file.stem}-proxy.fmu"
        return FMU(proxy_fmu_file)

    def _modify_start_values(self, variables_with_start_values: dict[str, ScalarVariable]):
        """Modify the start values of variables inside the FMUs modelDescription.xml."""

        logger.info(f"{self.file.name}: update start values of variables in modelDescription.xml")  # 2

        model_variables: MutableMapping[Any, Any] = self.model_description[
            find_key(self.model_description, "ModelVariables$")
        ]

        names_of_variables_with_start_values: list[str] = [
            variable.name for _, variable in variables_with_start_values.items()
        ]

        for model_variable_key, model_variable_properties in model_variables.items():
            model_variable_name: str = model_variable_properties["_attributes"]["name"]

            if model_variable_name in names_of_variables_with_start_values:
                variable_with_start_values = variables_with_start_values[model_variable_name]
                type_identifier = find_type_identifier_in_keys(model_variable_properties)
                type_key = find_key(model_variable_properties, f"{type_identifier}$")

                logger.info(
                    f"{self.file.name}: update start values for variable {model_variable_name}:\n"
                    f"\tstart:\t\t{variable_with_start_values.start}\n"
                    f"\tcausality:\t {variable_with_start_values.causality}\n"
                    f"\tvariability:\t{variable_with_start_values.variability}"
                )

                model_variables[model_variable_key][type_key]["_attributes"]["start"] = variable_with_start_values.start
                model_variables[model_variable_key]["_attributes"]["causality"] = variable_with_start_values.causality
                model_variables[model_variable_key]["_attributes"][
                    "variability"
                ] = variable_with_start_values.variability

        self._log_update_in_model_description()

    def _log_update_in_model_description(self, model_description: Union[CppDict, None] = None):
        model_description = model_description or self.model_description

        logger.info(f"{self.file.name}: update <fmiModelDescription description>")  # 2
        # Author
        old_author = model_description["_xmlOpts"]["_rootAttributes"]["author"]
        if platform.system() == "Linux":
            new_author = os.environ["USER"]
        else:
            new_author = os.environ["USERNAME"]
        model_description["_xmlOpts"]["_rootAttributes"]["author"] = new_author
        # DateAndTime
        old_date = model_description["_xmlOpts"]["_rootAttributes"]["generationDateAndTime"]
        new_date = str(datetime.now())
        model_description["_xmlOpts"]["_rootAttributes"]["generationDateAndTime"] = new_date
        # Log modifications in <fmiModelDescription description> attribute
        add_description_string = (
            f"\nmodified {date.today()}:\n"
            f"\tauthor {old_author} to {new_author}\n"
            f"\tgenerationDateAndTime {old_date} to {new_date}\n"
        )
        model_description["_xmlOpts"]["_rootAttributes"]["description"] += add_description_string

    # @TODO: Check when and where this method needs to be called. And why..
    #        CLAROS, 2022-05-24
    def _clean_solver_internal_variables(self, model_description: MutableMapping[Any, Any]):
        """Clean solver internal variables, such as '_iti_...'."""
        model_variables: Mapping[Any, Any] = model_description[find_key(model_description, "ModelVariables$")]
        model_name = model_description["_xmlOpts"]["_rootAttributes"]["modelName"]
        for model_variable_key in model_variables:
            if "_origin" in model_variables[model_variable_key]:
                model_variables[model_variable_key]["_origin"] = model_name



import logging
import os
import re
from pathlib import Path
from typing import Any, Dict, List, Union

from dictIO import DictReader, DictWriter
from dictIO.utils.counter import BorgCounter
from dictIO.utils.path import highest_common_root_folder, relative_path

from ospx.utils.dict import find_key, find_keys, find_type_identifier_in_keys

__ALL__ = ["OspSystemStructureImporter"]

logger = logging.getLogger(__name__)


class OspSystemStructureImporter:
    """Class providing methods to convert an existing
    OspSystemStructure.xml file to an ospx caseDict file.
    """

    @staticmethod
    def import_system_structure(
        system_structure_file: Union[str, os.PathLike[str]],
        enter_lib_source_as_relative_path: bool = False,
    ):
        """Import an OspSystemStructure.xml file and save it as an ospx caseDict file.

        Parameters
        ----------
        system_structure_file : Union[str, os.PathLike[str]]
            the OspSystemStructure.xml file to be imported
        enter_lib_source_as_relative_path : bool, optional
            whether lib_source shall be entered as relative path in the caseDict, by default False

        Raises
        ------
        FileNotFoundError
            if system_structure_file does not exist

        NotImplementedError
            if the OspSystemStructure contains connections of OSP-IS type 'SignalConnection' or 'SignalGroupConnection'.
            These connection types are not implemented yet in ospx.

        NotImplementedError
            if the OspSystemStructure contains connections with OSP-IS endpoint type 'Signal' or 'SignalGroup'.
            These endpoint types are not implemented yet in ospx.

        TypeError
            if the OspSystemStructure contains connections of an unknown type.

        TypeError
            if the OspSystemStructure contains connections with an unknown endpoint type.
        """

        # Make sure source_file argument is of type Path. If not, cast it to Path type.
        system_structure_file = (
            system_structure_file if isinstance(system_structure_file, Path) else Path(system_structure_file)
        )

        # Check whether system structure file exists
        if not system_structure_file.exists():
            logger.error(f"OspSystemStructureImporter: File {system_structure_file} not found.")
            raise FileNotFoundError(system_structure_file)

        if system_structure_file.suffix != ".xml":
            logger.error(f"OspSystemStructureImporter: File type {system_structure_file} not implemented yet.")
            return

        counter = BorgCounter()

        source_dict = DictReader.read(system_structure_file, comments=False)
        source_folder: Path = system_structure_file.resolve().parent.absolute()
        lib_source_folder: Path = source_folder  # setting to source_folder acts as fallback / default
        target_folder: Path = Path.cwd().absolute()

        # Main subdicts contained in systemStructure
        connections: Dict[str, Dict[Any, Any]] = {}
        components: Dict[str, Dict[Any, Any]] = {}

        # Connections
        # iterate over the connections first as they contain the variable and component names
        temp_connectors = {}
        if connections_key := find_key(source_dict, "Connections$"):
            for connection_type, connection_properties in source_dict[connections_key].items():
                connection_type: str = re.sub(r"(^\d{1,6}_)", "", connection_type)

                if connection_type not in {
                    "VariableConnection",
                    "VariableGroupConnection",
                }:
                    if connection_type in {"SignalConnection", "SignalGroupConnection"}:
                        msg: str = (
                            f"Import failed: {system_structure_file.name} contains a connection with OSP-IS connection type '{connection_type}'\n"
                            f"The support for connection type '{connection_type}' is not yet implemented in ospx."
                        )
                        logger.error(msg)
                        raise NotImplementedError(msg)
                    else:
                        msg: str = (
                            f"Import failed: {system_structure_file.name} contains a connection with unknown connection type '{connection_type}'\n"
                        )
                        logger.error(msg)
                        raise TypeError(msg)

                connection: Dict[str, Dict[Any, Any]] = {}
                connection_name: str = ""
                # following loop has range {0,1}
                for index, (endpoint_type, endpoint_properties) in enumerate(connection_properties.items()):
                    endpoint_type: str = re.sub(r"(^\d{1,6}_)", "", endpoint_type)

                    if endpoint_type not in {"Variable", "VariableGroup"}:
                        if endpoint_type in {"Signal", "SignalGroup"}:
                            msg: str = (
                                f"Import failed: {system_structure_file.name} contains a connection with OSP-IS endpoint type '{endpoint_type}'\n"
                                f"The support for endpoint type '{endpoint_type}' is not yet implemented in ospx."
                            )
                            logger.error(msg)
                            raise NotImplementedError(msg)
                        else:
                            msg: str = (
                                f"Import failed: {system_structure_file.name} contains a connection with unknown endpoint type '{endpoint_type}'\n"
                            )
                            logger.error(msg)
                            raise TypeError(msg)

                    component_name: str = endpoint_properties["_attributes"]["simulator"]
                    referenced_name: str = endpoint_properties["_attributes"]["name"]
                    # alternator for source <--> target (because there are always 2 entries in VariableConnection in always the same sequence)
                    endpoint_name: str = "source" if index % 2 == 0 else "target"
                    endpoint: dict[str, str] = {}
                    _connector_type: str = "output" if endpoint_name == "source" else "input"
                    _connector: dict[str, str] = {}
                    _connector_name: str = f"{component_name}_{referenced_name}"
                    if endpoint_type == "Variable":
                        endpoint = {
                            "component": component_name,
                            "variable": referenced_name,
                        }
                        _connector = {
                            "variable": referenced_name,
                            "type": _connector_type,
                        }
                    elif endpoint_type == "VariableGroup":
                        endpoint = {
                            "component": component_name,
                            "connector": _connector_name,
                        }
                        _connector = {
                            "variableGroup": referenced_name,
                            "type": _connector_type,
                        }
                    connection[endpoint_name] = endpoint
                    if not connection_name:
                        connection_name = component_name
                    else:
                        connection_name += f"_to_{component_name}"

                    # Save _connector in temp_connectors dict.
                    # (The variable and component information stored in these connectors
                    #  is later used to complete component properties)
                    temp_connectors[f"{counter():06d}_{component_name}"] = {_connector_name: _connector}
                # Save in connections dict
                if connection_name in connections:
                    suffix_number: int = 2
                    _connection_name: str = f"{connection_name}_{suffix_number:02d}"
                    while _connection_name in connections:
                        suffix_number += 1
                        _connection_name = f"{connection_name}_{suffix_number:02d}"
                    connection_name = _connection_name
                connections[connection_name] = connection
        # Simulators (=Components)
        if simulators_key := find_key(source_dict, "Simulators$"):
            # Determine the highest common root folder among all FMU's.
            # This will be used as libSource folder.
            fmu_folders: List[Path] = []
            for simulator_properties in source_dict[simulators_key].values():
                fmu_name: str = simulator_properties["_attributes"]["source"]
                fmu_file: Path = Path(fmu_name)
                fmu_folder: Path
                if fmu_file.is_absolute():
                    fmu_folder = fmu_file.resolve().parent.absolute()
                else:
                    fmu_folder = (source_folder / fmu_file).resolve().parent.absolute()
                fmu_folders.append(fmu_folder)
            if fmu_folders:
                lib_source_folder = highest_common_root_folder(fmu_folders)

            for simulator_properties in source_dict[simulators_key].values():
                # Component
                component_name = simulator_properties["_attributes"]["name"]
                # Connectors
                component_connectors: Dict[str, Dict[Any, Any]] = {}
                for temp_connector_key, connector in temp_connectors.items():
                    if component_name in temp_connector_key:
                        component_connectors |= connector
                # FMU
                fmu_name: str = simulator_properties["_attributes"]["source"]
                fmu_file: Path = Path(fmu_name)
                if fmu_file.is_absolute():
                    fmu_file = fmu_file.resolve()
                else:
                    fmu_file = (source_folder / fmu_file).resolve()
                fmu_file_relative_to_lib_source: Path = relative_path(lib_source_folder, fmu_file)
                # Step Size
                step_size: Union[float, None] = None
                if "stepSize" in simulator_properties["_attributes"]:
                    step_size = float(simulator_properties["_attributes"]["stepSize"])
                    # Initial values
                component_initial_values: Dict[str, Dict[Any, Any]] = {}
                if initial_values_key := find_key(simulator_properties, "InitialValues$"):
                    initial_values = simulator_properties[initial_values_key]
                    if initial_value_keys := find_keys(initial_values, "InitialValue$"):
                        for initial_value_key in initial_value_keys:
                            initial_value = initial_values[initial_value_key]
                            if data_type := find_type_identifier_in_keys(initial_value):
                                type_key = find_key(initial_value, f"{data_type}$")
                                if not type_key:
                                    continue
                                _type: str = re.sub(r"(^\d{1,6}_)", "", type_key)
                                referenced_name = initial_value["_attributes"]["variable"]
                                value: Union[float, int, bool, str]
                                if _type == "Boolean":
                                    value = bool(initial_value[type_key]["_attributes"]["value"])
                                elif _type == "Integer":
                                    value = int(initial_value[type_key]["_attributes"]["value"])
                                elif _type == "Real":
                                    value = float(initial_value[type_key]["_attributes"]["value"])
                                else:
                                    value = initial_value[type_key]["_attributes"]["value"]
                                component_initial_values |= {referenced_name: {"start": value}}
                                # Assemble component
                component: Dict[str, Union[Dict[Any, Any], str, float, Path]] = {
                    "connectors": component_connectors,
                    "fmu": fmu_file_relative_to_lib_source,
                }
                if step_size:
                    component["stepSize"] = step_size
                if component_initial_values:
                    component["initialize"] = component_initial_values
                    # Save in components dict
                components[component_name] = component

        # System Structure
        system_structure: Dict[str, Dict[str, Any]] = {
            "connections": connections,
            "components": components,
        }

        # Global Settings
        # 1: Defaults
        simulation: Dict[str, Any] = {
            "name": system_structure_file.stem,
            "startTime": 0.0,
            "baseStepSize": 0.01,
            "algorithm": "fixedStep",
        }
        # 2: Overwrite defaults with values from source dict, where existing
        if "_attributes" in source_dict:
            attributes = source_dict["_attributes"]
            if "StartTime" in attributes:
                simulation["startTime"] = attributes["StartTime"]
            if "BaseStepSize" in attributes:
                simulation["baseStepSize"] = attributes["BaseStepSize"]
            if "Algorithm" in attributes:
                simulation["algorithm"] = attributes["Algorithm"]

        # Environment
        environment: dict[str, Path] = {}
        if enter_lib_source_as_relative_path:
            try:
                environment["libSource"] = relative_path(
                    from_path=target_folder,
                    to_path=lib_source_folder,
                )
            except ValueError:
                msg = (
                    "Resolving relative path from target folder to libSource folder failed.\n"
                    "Absolute path for libSource is used instead."
                )
                logger.warning(msg)
                environment["libSource"] = lib_source_folder
        else:
            environment["libSource"] = lib_source_folder

        # Assemble case dict
        case_dict = {
            "_environment": environment,
            "systemStructure": system_structure,
            "run": {
                "simulation": simulation,
            },
        }

        source_file_name = source_dict.name.replace(".", "_")
        target_file = Path.cwd() / f"caseDict_imported_from_{source_file_name}"

        DictWriter.write(case_dict, target_file, mode="w")

        return



import logging
import os
import re
from pathlib import Path
from shutil import copyfile
from tempfile import mkstemp
from typing import Tuple, Union
from zipfile import ZIP_DEFLATED, ZipFile

logger = logging.getLogger(__name__)


def read_file_content_from_zip(zip_file: Path, file_name: str) -> Union[str, None]:
    """
    belongs to zip functions
    read a single file.
    """
    file_handle, temp_name = mkstemp(dir=zip_file.parent)
    file_content = None
    try:
        with ZipFile(zip_file, "r") as zip_read:
            for item in zip_read.infolist():
                if re.search(file_name, item.filename):
                    file_content = str(zip_read.read(item.filename).decode("utf-8"))
                    break
    except Exception:
        logger.exception("misc.zip.read_file_content_from_zip failed")
    finally:
        os.close(file_handle)
        os.remove(temp_name)
    return file_content


def rename_file_in_zip(zip_file: Path, file_name: str, new_file_name: str) -> Union[ZipFile, None]:
    """
    belongs to zip functions
    rename files.
    """
    file_handle, temp_name = mkstemp(dir=zip_file.parent)
    updated_zip_file = None
    try:
        with ZipFile(zip_file, "r") as zip_read:
            with ZipFile(temp_name, "w") as zip_write:
                for item in zip_read.infolist():
                    if item.filename != file_name:
                        data = zip_read.read(item.filename)
                    else:
                        data = zip_read.read(item.filename)
                        item.filename = new_file_name
                    zip_write.writestr(item, data)
        _ = copyfile(temp_name, zip_file)

        updated_zip_file = ZipFile(zip_file, mode="a")

    except Exception:
        logger.exception("misc.zip.rename_file_in_zip failed")
    finally:
        os.close(file_handle)
        os.remove(temp_name)

    return updated_zip_file


def remove_files_from_zip(zip_file: Path, *file_names: str) -> Union[ZipFile, None]:
    """
    belongs to zip functions
    remove files.
    """
    file_handle, temp_name = mkstemp(dir=zip_file.parent)
    updated_zip_file = None
    try:
        with ZipFile(zip_file, "r") as zip_read:
            with ZipFile(temp_name, "w") as zip_write:
                for item in zip_read.infolist():
                    if item.filename not in file_names:
                        data = zip_read.read(item.filename)
                        zip_write.writestr(item, data)

        _ = copyfile(temp_name, zip_file)

        updated_zip_file = ZipFile(zip_file, mode="a")

    except Exception:
        logger.exception("misc.zip.remove_files_from_zip failed")

    finally:
        os.close(file_handle)
        os.remove(temp_name)

    return updated_zip_file


def add_file_content_to_zip(zip_file: Path, file_name: str, file_content: str) -> Union[ZipFile, None]:
    """
    belongs to zip functions
    add a single file and its ascii content.
    """
    file_handle, temp_name = mkstemp(dir=zip_file.parent)
    updated_zip_file = None
    try:
        with ZipFile(zip_file, "a") as zip_write:
            zip_write.writestr(file_name, file_content, compress_type=ZIP_DEFLATED)

        updated_zip_file = ZipFile(zip_file, mode="a")

    except Exception:
        logger.exception("misc.zip.add_file_content_to_zip failed")
    finally:
        os.close(file_handle)
        os.remove(temp_name)

    return updated_zip_file


def substitute_text_in_zip(
    zip_file: Path, file_name_pattern: str = "", subst: Tuple[str, str] = ("", "")
) -> Union[ZipFile, None]:
    """
    belongs to zip functions
    substitutes a given string in all files matching the passed file name pattern.
    """

    file_handle, temp_name = mkstemp(dir=zip_file.parent)
    updated_zip_file = None
    try:
        with ZipFile(zip_file, "r") as zip_read:
            with ZipFile(temp_name, "w") as zip_write:
                zip_write.comment = zip_read.comment  # preserve the comment
                for item in zip_read.infolist():
                    if not re.search(file_name_pattern, item.filename):
                        zip_write.writestr(item, zip_read.read(item.filename))
                    else:
                        temp = zip_read.read(item.filename)
                        source = (re.findall(subst[0], str(temp)))[0]
                        if not str(source):
                            logger.warning(f'substitution source is empty:\'{" ".join(source)}\'')
                        temp = temp.replace(bytes(source, "utf-8"), bytes(subst[1], "utf-8"))
                        zip_write.writestr(item, temp)

        updated_zip_file = ZipFile(zip_file, mode="a")

    except Exception:
        logger.exception("misc.zip.substitute_text_in_zip failed")
    finally:
        os.close(file_handle)
        os.remove(temp_name)

    return updated_zip_file


def update_file_content_in_zip(zip_file: Path, file_name: str, file_content: str) -> Union[ZipFile, None]:
    """
    belongs to zip functions
    updates the ascii content of a single file.
    """

    file_handle, temp_name = mkstemp(dir=zip_file.parent)
    updated_zip_file = None
    try:
        with ZipFile(zip_file, "r") as zip_read:
            with ZipFile(temp_name, "w") as zip_write:
                zip_write.comment = zip_read.comment  # preserve the comment
                for item in zip_read.infolist():
                    if item.filename != file_name:
                        zip_write.writestr(item, zip_read.read(item.filename))

        with ZipFile(zip_file, mode="a", compression=ZIP_DEFLATED) as zf:
            # zf.writestr(contentFile, '<?xml version="1.0" encoding="UTF-8"?>\n'+data.decode('utf-8'))
            zf.writestr(file_name, file_content)

        updated_zip_file = ZipFile(zip_file, mode="a")

    except Exception:
        logger.exception("misc.zip.update_file_content_in_zip failed")
    finally:
        os.close(file_handle)
        os.remove(temp_name)

    return updated_zip_file



from pathlib import Path

import pytest

from ospx.fmi.fmu import FMU


def test_conftest_create_test_fmu():
    pass


def test_fmu_instantiation():
    # Prepare
    fmu_file: Path = Path("test_fmu.fmu")
    # Execute
    fmu = FMU(fmu_file)
    # Assert
    assert isinstance(fmu, FMU)


@pytest.fixture()
def test_fmu() -> FMU:
    fmu_file: Path = Path("test_fmu.fmu")
    return FMU(fmu_file)


def test_fmu_variables_number(test_fmu: FMU):
    assert len(test_fmu.variables) == 34


def test_fmu_units_number(test_fmu: FMU):
    assert len(test_fmu.units) == 0


def test_fmu_variables_fmi_data_type(test_fmu: FMU):
    assert test_fmu.variables["Variable_1_IN_Real"].data_type == "Real"
    assert isinstance(test_fmu.variables["Variable_1_IN_Real"].start, float)
    assert test_fmu.variables["Variable_2_IN_Integer"].data_type == "Integer"
    assert test_fmu.variables["Variable_3_IN_Bool"].data_type == "Boolean"
    assert test_fmu.variables["Variable_4_OUT_Real"].data_type == "Real"
    assert test_fmu.variables["Variable_5_OUT_Integer"].data_type == "Integer"
    assert test_fmu.variables["Variable_6_OUT_Bool"].data_type == "Boolean"


def test_fmu_variables_start_value(test_fmu: FMU):
    assert test_fmu.variables["Vector_1_IN[0]"].start == 10.0
    assert test_fmu.variables["Vector_1_IN[1]"].start == 11.0
    assert test_fmu.variables["Vector_1_IN[2]"].start == 12.0


# def test_fmu():
# Prepare

# Execute

# Assert



import logging
from typing import Union

from ospx.component import Component
from ospx.connector import Connector
from ospx.fmi.variable import ScalarVariable

__ALL__ = ["Endpoint", "Connection"]

logger = logging.getLogger(__name__)


class Endpoint:
    """Endpoints relate each side of a connection to distinct variables or connectors.

    An endpoint is for a connection what a connector is for a component.
    A connection has two endpoints, one for each side.
    Connection endpoints are hence the 'counterparts' to component's connectors.
    """

    def __init__(
        self,
        component: Component,
        connector: Union[Connector, None] = None,
        variable: Union[ScalarVariable, None] = None,
    ):
        self.component: Component = component
        self._connector: Union[Connector, None] = None
        self._variable: Union[ScalarVariable, None] = None
        if connector:
            self.connector = connector
        if variable:
            self.variable = variable

    @property
    def connector(self) -> Union[Connector, None]:
        """Returns the connector this endpoint refers to, if defined.

        Returns
        -------
        Union[Connector, None]
            the connector, if defined. Otherwise None.
        """
        return self._connector

    @connector.setter
    def connector(self, connector: Connector):
        """Set the connector this endpoint shall refer to.

        Parameters
        ----------
        connector : Connector
            the connector
        """
        if self._variable:
            msg = (
                f"Inconsistency: Connection endpoint defines both connector and variable.\n"
                f"connector: {connector.name}\nvariable: {self._variable.name}\n"
                "connector is used. variable is omitted."
            )
            logger.warning(msg)
            self._variable = None
        self._connector = connector

    @property
    def variable(self) -> Union[ScalarVariable, None]:
        """Returns the scalar variable this endpoint refers to, if defined.

        Returns
        -------
        Union[ScalarVariable, None]
            the scalar variable, if defined. Otherwise None.
        """
        return self._variable

    @variable.setter
    def variable(self, variable: ScalarVariable):
        """Set the scalar variable this endpoint shall refer to.

        Parameters
        ----------
        variable : ScalarVariable
            the scalar variable
        """
        if self._connector:
            msg = (
                f"Inconsistency: Connection endpoint defines both connector and variable.\n"
                f"connector: {self._connector.name}\nvariable: {variable.name}\n"
                "connector is omitted. variable is used."
            )
            logger.warning(msg)
            self._connector = None
        self._variable = variable

    @property
    def variable_name(self) -> str:
        """Returns the name of the scalar variable this endpoint refers to.

        Returns
        -------
        str
            the name of the scalar variable.
        """
        if self._connector:
            return self._connector.variable_name
        elif self._variable:
            return self._variable.name
        else:
            return "UNKNOWN"

    @property
    def is_valid(self) -> bool:
        """Consistency check. Returns True if endpoint is defined and valid.

        Returns
        -------
        bool
            True if valid. Otherwise False.
        """
        return bool(self.component and (self.connector or self.variable))


class Connection:
    """A connection is the primary artefact to connect outputs and inputs of componoents in a system.

    A connection connects an output connector of one component with an input connector of another component.
    """

    def __init__(
        self,
        name: str,
        source_endpoint: Endpoint,
        target_endpoint: Endpoint,
    ):
        self.name: str = name
        self.source_endpoint: Endpoint = source_endpoint
        self.target_endpoint: Endpoint = target_endpoint

    @property
    def is_variable_connection(self) -> bool:
        """Returns True if connection is a single variable connection.

        Returns
        -------
        bool
            True if single variable connection. Otherwise False.
        """
        if not self.source_endpoint:
            return False
        if self.source_endpoint.variable:
            return True
        if self.source_endpoint.connector:
            return self.source_endpoint.connector.is_single_connector
        return False

    @property
    def is_variable_group_connection(self) -> bool:
        """Returns True if connection is a variable group connection.

        Returns
        -------
        bool
            True if variable group connection. Otherwise False.
        """
        if not self.source_endpoint:
            return False
        if self.source_endpoint.variable:
            return False
        if self.source_endpoint.connector:
            return self.source_endpoint.connector.is_group_connector
        return False

    @property
    def is_valid(self) -> bool:
        """Consistency check. Returns True if connection is found fully defined and valid.

        Returns
        -------
        bool
            True if valid. Otherwise False.
        """
        if not (self.source_endpoint.is_valid and self.target_endpoint.is_valid):
            return False
        if self.source_endpoint.variable and self.target_endpoint.variable:
            return True
        if self.source_endpoint.connector and self.target_endpoint.connector:
            _both_are_single: bool = (
                self.source_endpoint.connector.is_single_connector
                and self.target_endpoint.connector.is_single_connector
            )
            _both_are_group: bool = (
                self.source_endpoint.connector.is_group_connector and self.target_endpoint.connector.is_group_connector
            )
            return _both_are_single or _both_are_group
        elif self.source_endpoint.connector:
            return self.source_endpoint.connector.is_single_connector
        elif self.target_endpoint.connector:
            return self.target_endpoint.connector.is_single_connector
        return False



import logging
from dataclasses import dataclass
from typing import Union

__ALL__ = ["Experiment"]

logger = logging.getLogger(__name__)


@dataclass()
class Experiment:
    """Data class for the DefaultExperiment element inside fmi 2.0 ModelDescription.

    See https://github.com/modelica/fmi-standard/blob/v2.0.x/schema/fmi2ModelDescription.xsd
    """

    start_time: Union[float, None] = None
    stop_time: Union[float, None] = None
    tolerance: Union[float, None] = None
    step_size: Union[float, None] = None



import logging
from typing import Any, MutableMapping, Union

from ospx import Component, Connection, Connector, Endpoint
from ospx.fmi import FMU, ScalarVariable, Unit

__ALL__ = ["System"]

logger = logging.getLogger(__name__)


class System:
    """The system structure describes the topology of the co-simulated system.

    A system structure can contain an arbitrary number of components.
    Components can be connected through connections.
    Connections relate a source endpoint with a target endpoint.
    Both component variables and component connectors can be used as endpoints in a connection.
    """

    def __init__(self, properties: MutableMapping[Any, Any]):
        self._components: dict[str, Component] = {}
        self._connections: dict[str, Connection] = {}
        self._read_components(properties)
        self._read_connections(properties)

    @property
    def fmus(self) -> dict[str, FMU]:
        """Returns a dict with all FMUs referenced by components contained in the system.

        Returns
        -------
        dict[str, FMU]
            dict with all FMUs
        """
        return {component.fmu.file.name: component.fmu for component in self.components.values() if component.fmu}

    @property
    def components(self) -> dict[str, Component]:
        """Returns a dict with all components contained in the system.

        Returns
        -------
        dict[str, Component]
            dict with all components
        """
        return self._components

    @property
    def connections(self) -> dict[str, Connection]:
        """Returns a dict with all connections defined in the system.

        Returns
        -------
        dict[str, Connection]
            dict with all connections
        """
        return self._connections

    @property
    def units(self) -> dict[str, Unit]:
        """Returns a combined dict with all units
        from all components contained in the system.

        Returns
        -------
        dict[str, Unit]
            dict with all units from all components
        """
        units: dict[str, Unit] = {}
        for component in self.components.values():
            if component.units:
                units |= component.units
        return units

    @property
    def connectors(self) -> dict[str, Connector]:
        """Returns a combined dict with all connectors
        from all components contained in the system.

        Returns
        -------
        dict[str, Connector]
            dict with all connectors from all components
        """
        connectors: dict[str, Connector] = {}
        for component in self.components.values():
            if component.connectors:
                connectors |= component.connectors
        return connectors

    @property
    def variables(self) -> dict[str, ScalarVariable]:
        """Returns a combined dict with all scalar variables
        from all components contained in the system.

        Returns
        -------
        dict[str, ScalarVariable]
            dict with all scalar variables from all components
        """
        variables: dict[str, ScalarVariable] = {}
        for component in self.components.values():
            if component.variables:
                variables |= component.variables
        return variables

    def _read_components(self, properties: MutableMapping[Any, Any]):
        """Read components from (case dict) properties."""
        logger.info("read components from case dict")
        self._components.clear()
        if "components" not in properties:
            return
        for component_name, component_properties in properties["components"].items():
            component = Component(component_name, component_properties)
            self._components[component.name] = component

    def _read_connections(self, properties: MutableMapping[Any, Any]):
        """Read connections from (case dict) properties."""
        logger.info("read connections from case dict")
        self._connections.clear()
        if "connections" not in properties:
            return
        for connection_name, connection_properties in properties["connections"].items():
            source_endpoint: Union[Endpoint, None] = None
            target_endpoint: Union[Endpoint, None] = None
            if "source" in connection_properties:
                source_endpoint = self._read_endpoint(connection_properties["source"])
            if "target" in connection_properties:
                target_endpoint = self._read_endpoint(connection_properties["target"])
            if source_endpoint and target_endpoint:
                connection = Connection(
                    name=connection_name,
                    source_endpoint=source_endpoint,
                    target_endpoint=target_endpoint,
                )
                self._connections[connection.name] = connection
            else:
                logger.error(
                    f"connection {connection_name}: connection could not be resolved. Please recheck connection properties in case dict."
                )
        return

    def _read_endpoint(self, properties: MutableMapping[Any, Any]) -> Union[Endpoint, None]:
        if "component" not in properties:
            return None
        component: Union[Component, None] = None
        connector: Union[Connector, None] = None
        variable: Union[ScalarVariable, None] = None

        component_name: str = properties["component"]
        if component_name in self.components:
            component = self.components[component_name]

        if "connector" in properties:
            connector_name = properties["connector"]
            if component and connector_name in component.connectors:
                connector = component.connectors[connector_name]
            else:
                for _, c in self.components.items():
                    if connector_name in c.connectors:
                        component = c
                        connector = c.connectors[connector_name]
                        break

        if "variable" in properties:
            variable_name = properties["variable"]
            if component and variable_name in component.variables:
                variable = component.variables[variable_name]

        if not component:
            return None

        if connector or variable:
            endpoint: Endpoint = Endpoint(
                component=component,
                connector=connector,
                variable=variable,
            )
            return endpoint if endpoint.is_valid else None

        return None



import logging
from dataclasses import dataclass
from typing import Union

__ALL__ = ["Simulation"]

logger = logging.getLogger(__name__)


@dataclass()
class Simulation:
    """Data class holding the attributes of the 'simulation' element inside OspSystemStructure.xml."""

    name: Union[str, None] = None
    start_time: Union[float, None] = None
    stop_time: Union[float, None] = None
    base_step_size: Union[float, None] = None
    _algorithm: Union[str, None] = None

    @property
    def algorithm(self) -> Union[str, None]:  # noqa: D102
        return self._algorithm

    @algorithm.setter
    def algorithm(self, value: str):  # noqa: D102
        valid_values: list[str] = [
            "fixedStep",
        ]
        if value not in valid_values:
            logger.error(f"variable {self.name}: algorithm value '{value}' is invalid.")
            return
        self._algorithm = value
        return



# pyright: reportUnknownMemberType=false
# pyright: reportArgumentType=false
# pyright: reportCallIssue=false

import contextlib
import logging
import os
import re
from math import sqrt as sqrt
from pathlib import Path
from typing import Any, Dict, List, MutableMapping, MutableSequence, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from dictIO import DictReader, DictWriter
from matplotlib import cm
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from numpy import ndarray
from pandas import DataFrame

from ospx.utils.plotting import create_meta_dict, save_figure

logger = logging.getLogger(__name__)


class CosimWatcher:
    """CosimWatcher allows to monitor a running simulation,
    plot trends and dump simulation results into a resultDict file.
    """

    def __init__(
        self,
        csv_file_names: MutableSequence[str],
        skip_values: int,
        latest_values: int,
        scale_factor: float,
        timeline_data: bool,
    ):
        self.watch_dict_file: Union[Path, None] = None
        self.watch_dict: MutableMapping[Any, Any] = {}
        self.csv_file_names: MutableSequence[str] = csv_file_names
        self.title: str = "watch"
        self.delimiter: str = ","  # default
        self.data_sources: Dict[str, Dict[str, Union[List[int], List[str], int, str]]] = {}
        self.results_dir: str = "results"
        self.number_of_columns: int = 3
        self.number_of_subplots: int = 0
        self.skip_values: int = skip_values
        self.latest_values: int = latest_values
        self.scale_factor: float = scale_factor
        self.timeline_data: bool = timeline_data
        self.figure: Figure
        self.terminate: bool = False
        self.max_row: int = 0
        return

    def read_watch_dict(self, watch_dict_file: Union[str, os.PathLike[str]]):
        """Read watchDict file.

        The watchDict file contains the parameters to be plotted.

        Parameters
        ----------
        watch_dict_file : Union[str, os.PathLike[str]]
            watchDict file. Contains the parameters to be plotted.

        Raises
        ------
        FileNotFoundError
            if watch_dict_file does not exist
        """

        # Make sure watch_dict_file argument is of type Path. If not, cast it to Path type.
        watch_dict_file = watch_dict_file if isinstance(watch_dict_file, Path) else Path(watch_dict_file)
        if not watch_dict_file.exists():
            logger.error(f"CosimWatcher: File {watch_dict_file} not found.")
            raise FileNotFoundError(watch_dict_file)

        logger.info(f"Configure CosimWatcher with {watch_dict_file}..")

        self.watch_dict_file = watch_dict_file

        self.watch_dict = DictReader.read(Path(self.watch_dict_file), comments=False)

        # read datasources, if available.
        # normally this part should be written by ospCaseBuilder entirely
        self.data_sources = self.watch_dict["datasources"]

        if "delimiter" in self.watch_dict:
            self.delimiter = self.watch_dict["delimiter"]

        if "simulation" in self.watch_dict:
            self.title = f"{self.watch_dict_file.name}-{self.watch_dict['simulation']['name']}"

        self._define_data_source_properties_for_plotting()
        return

    def plot(self, converge: bool = False):
        """Plot trends.

        Plotting + convergence checker (future task)

        Parameters
        ----------
        converge : bool, optional
            if True, convergence is checked, by default False
        """

        self._initialize_plot()

        if converge:
            terminate_loops = 0
            max_no_change_loops = 4
        else:
            terminate_loops = 10
            max_no_change_loops = 0

        df_row_size = 0

        while True:  # do as long as not interrupted
            df = self._read_csv_files_into_dataframe()

            # cumulate counter for termination if no changes
            if df_row_size == len(df):
                terminate_loops += 1
            else:
                terminate_loops = 0

            df_row_size = len(df)
            df_col_size = (
                len(list(df)) - 1
            )  # reduced by one because 1st col is time column, frl: check multiple datasources

            # axs = [None for x in range(df_col_size)]
            axs: MutableSequence[Axes] = []
            plot: Axes
            time_key = list(df)[0]
            for index in range(df_col_size):
                current_key = list(df)[index + 1]  # 0 is time column and thus removed, frl: check multiple dataframes

                plot = self.figure.add_subplot(self.max_row, self.number_of_columns, index + 1)

                try:
                    _ = plot.plot(
                        time_key,
                        current_key,
                        linewidth=2,
                        color=cm.get_cmap("gist_rainbow")(index / self.number_of_subplots),
                        data=df[[time_key, current_key]],
                    )
                except (TypeError, ValueError):
                    pass
                except Exception as e:
                    logger.exception(e)

                # subplot.set_title(currentKey,  fontsize=10)
                _ = plot.grid(color="#66aa88", linestyle="--")
                _ = plot.xaxis.set_tick_params(labelsize=8)
                _ = plot.yaxis.set_tick_params(labelsize=8)
                _ = plot.legend(fontsize=8)
                axs.append(plot)
                # if isinstance(plot, Axes):
                #     axs.append(plot)
                # else:
                #     raise TypeError(
                #         f"CosimWatcher.plot(): plot is of type {type(plot)}. Expected type was matplotlib.axes.Axes ."
                #     )

            _ = self.figure.suptitle(self.title)

            if converge:
                plt.show(block=False)
                plt.pause(3)

            if terminate_loops >= max_no_change_loops:
                save_figure(
                    self.figure,
                    extension="png",
                    path=self.results_dir,
                    title=self.title,
                    meta_dict=create_meta_dict(self.title),
                )
                break
            plt.clf()

            # @TODO: Implement keypress for termination

        return

    def dump(self):
        """Write dataframe to dump."""

        df = self._read_csv_files_into_dataframe()

        result_dict = {}
        for header in list(df):
            values: ndarray[Any, Any] = df[header].dropna().to_numpy()
            _first_value: Any = values[0]
            _last_value: Any = values[-1]
            _mean: Union[float, str] = "None"
            _stddev: Union[float, str] = "None"
            _min: Union[float, str] = "None"
            _max: Union[float, str] = "None"
            with contextlib.suppress(TypeError):
                _mean = np.mean(values)  # type: ignore
            with contextlib.suppress(TypeError):
                _stddev = np.std(values)  # type: ignore
            with contextlib.suppress(TypeError):
                _min = np.min(values)
            with contextlib.suppress(TypeError):
                _max = np.max(values)
            result_dict[header] = {
                "latestValue": _last_value,
                "firstValue": _first_value,
                "mean": _mean,
                "stdev": _stddev,
                "min": _min,
                "max": _max,
            }
            if self.timeline_data:
                result_dict[header].update({"values": values})

        # debug
        # result_dict.update({'_datasources':self.data_sources})
        result_dict_name = "-".join([self.title, "resultDict"])

        target_file_path = Path.cwd() / self.results_dir / result_dict_name
        DictWriter.write(result_dict, target_file_path, mode="w")

        dump_dict_name = "-".join([self.title, "dataFrame.dump"])
        target_file_path = Path.cwd() / self.results_dir / dump_dict_name
        df.to_pickle(str(target_file_path.absolute()), compression="gzip")
        return

    def _define_data_source_properties_for_plotting(self):
        """Details out the properties of all data sources for plotting.

        Details out the properties of all data source, making sure they contain the following fields required for plotting
            - file name
            - column names (= variable names)
        """

        # pattern = re.compile(r"(^#|\s+\[.*?\]$)")
        pattern = re.compile(
            r"(^#{0,2}\s*|\s+\[.*?\]$)"
        )  # frl 2023-11-07 remove all leading #'s and spaces and all trailing [.*]'s

        for (
            data_source_name,
            data_source_properties,
        ) in self.data_sources.items():  # loop over all data sources
            for csv_file_name in self.csv_file_names:
                if re.match(data_source_name, csv_file_name):  # find the correct csv file
                    data_source_properties.update({"csvFile": csv_file_name})

                    # extract the header row from the csv file to determine the variable names
                    data_header: List[str] = []
                    with open(csv_file_name, "r") as f:
                        data_header = f.readline().strip().split(self.delimiter)
                    if not data_header:
                        continue

                    time_column: int = 0  # frl 2023-11-07 default first column
                    if "timeColumn" in data_source_properties and isinstance(data_source_properties["timeColumn"], int):
                        time_column = data_source_properties["timeColumn"]

                    _time_name: str = data_header[time_column]
                    data_source_properties.update({"timeName": _time_name})
                    _display_time_name: str = pattern.sub("", _time_name)
                    data_source_properties.update({"displayTimeName": _display_time_name})

                    data_columns: List[int] = []
                    # read_only_shortlisted_columns: bool = False #flr 2023-11-07 greedy approach needs to be updated on demand

                    read_only_shortlisted_columns = "dataColumns" in data_source_properties
                    if read_only_shortlisted_columns and (
                        "dataColumns" in data_source_properties
                        and isinstance(data_source_properties["dataColumns"], List)
                    ):
                        data_columns = [int(col) for col in data_source_properties["dataColumns"]]
                    # else: frl 2023-11-07 simx heritage?
                    #    # if columns were not explicitely specified in watch dict:
                    #    # Read all columns except settings.
                    #    columns.extend(
                    #        index
                    #        for index, col_name in enumerate(data_header)
                    #        if not re.match(r"^(settings)", col_name)
                    #    )

                    _column_names: List[str] = [data_header[column] for column in data_columns]
                    data_source_properties.update({"colNames": _column_names})
                    _display_column_names: List[str] = [pattern.sub("", col_name) for col_name in _column_names]
                    # _display_column_names = ["Time", "StepCount"] + [
                    _display_column_names = [
                        data_source_name + "|" + col_name
                        for col_name in _display_column_names
                        # if col_name not in ["Time", "StepCount"] frl 2023-11-07
                    ]

                    data_source_properties.update({"displayColNames": _display_column_names})
                    data_source_properties.update({"xColumn": time_column})
                    data_source_properties.update({"yColumns": data_columns})

        return

    def _initialize_plot(self):
        """Initialize the plot.

        Collects data and sets plot header line
        """
        self.figure = plt.figure(figsize=(16 * self.scale_factor, 9 * self.scale_factor), dpi=150)
        # self.fig.tight_layout()  # constraint_layout()
        _ = self.figure.subplots_adjust(
            left=0.1,
            bottom=0.05,
            right=0.95,
            top=0.9,
            wspace=0.2,
            hspace=0.2,
        )
        self.terminate = False

        df = (
            self._read_csv_files_into_dataframe()
        )  # do it once to find the number of respective columns of all datasources
        self.number_of_subplots = (
            len(list(df)) - 1
        )  # one of the columns is the abscissa, frl: check if this works for multiple datasources and merged time columns

        self.number_of_columns = int(sqrt(self.number_of_subplots - 1)) + 1
        self.max_row = int(self.number_of_subplots / self.number_of_columns - 0.1) + 1
        return

    def _read_csv_files_into_dataframe(self) -> DataFrame:
        """Read all csv files into one joint Pandas dataframe.

        Read all csv files (=all data sources, one csv file per data source) into one joint Pandas dataframe.
        The returned dataframe hence contains the data of all datas ources.
        This dataframe can then be used for plotting and to dump a pickle.

        Returns
        -------
        pandas.core.frame.DataFrame
            Pandas dataframe containing the data of all csv files
        """

        df_all_data_sources = pd.DataFrame()  # initialize empty df

        for _, data_source_properties in self.data_sources.items():
            # mapping dict for display column names
            column_name_to_display_column_name_mapping: Dict[str, str] = dict(
                zip(
                    [data_source_properties["timeName"]] + data_source_properties["colNames"],  # type: ignore
                    [data_source_properties["displayTimeName"]] + data_source_properties["displayColNames"],  # type: ignore
                )
            )

            _column_names: List[str] = []
            if "colNames" in data_source_properties and isinstance(data_source_properties["colNames"], List):
                _column_names = [str(col_name) for col_name in data_source_properties["colNames"]]
            if "timeName" in data_source_properties and isinstance(data_source_properties["timeName"], str):
                _column_names = [data_source_properties["timeName"]] + _column_names

            if "csvFile" in data_source_properties and isinstance(data_source_properties["csvFile"], str):
                df_single_data_source: DataFrame
                df_single_data_source = pd.read_csv(
                    Path(data_source_properties["csvFile"]),
                    usecols=_column_names,
                )

                df_single_data_source = df_single_data_source.rename(columns=column_name_to_display_column_name_mapping)

                if df_all_data_sources.empty:
                    # first df inherit all columns from single df
                    df_all_data_sources = df_single_data_source
                else:
                    # all subsequent merge row-wise by time column,
                    # ignoring index
                    # (after setting individual time steps for each individual component)

                    # concatenate column-wise
                    # df_all_data_sources = pd.concat([df_all_data_sources, df_single_data_source], axis=1)

                    # df_all_data_sources = pd.concat([df_all_data_sources, df_single_data_source], ignore_index=True)
                    df_all_data_sources = pd.concat(
                        [df_all_data_sources, df_single_data_source]
                    )  # frl check for duplicated timeName columns for multiple datasources

                    # potential solution
                    # interpolating non-matching time data
                    # otherwise should component-wise dataframes do a better job
                    # bypass StepCound yielding in big holes, not plotted by mpl.
                    # df_all_data_sources = pd.merge_asof(
                    #    df_all_data_sources,
                    #    df_single_data_source,
                    #    on = 'Time',
                    #    by = 'StepCount',
                    #    direction = 'nearest',
                    #    #tolerance = pd.Timedelta('1ms')
                    # )

        # find latest common start point for skip and latest
        # consider skipping negative values due to wrong inputs
        start: int = 0
        if df_all_data_sources.shape[0] - self.skip_values < 0:  # safety
            logger.error(f"there will be no data, consider adjusting --skip: {self.skip_values}")
            # cases
        if self.skip_values > 0 and self.latest_values > 0:
            start = max(self.skip_values, df_all_data_sources.shape[0] - self.latest_values)
        elif self.skip_values > 0 and self.latest_values == 0:
            start = self.skip_values
        elif self.latest_values > 0 and self.skip_values == 0:
            start = df_all_data_sources.shape[0] - self.latest_values
        else:
            start = 0

        # if skip latest n steps is to be implemented, no changes to start, but an additional command option is required
        length: int = df_all_data_sources.shape[0]

        return df_all_data_sources.iloc[start:length, :]

    def _determine_optimum_screen_size(self):
        """Determine the optimum screen size."""
        # Opening and closing of window may be deprecated when a better solution is found
        mgr = plt.get_current_fig_manager()
        if mgr is None:
            return
        mgr.full_screen_toggle()
        self.screenSize = (mgr.canvas.width(), mgr.canvas.height())  # type: ignore
        mgr.window.close()  # type: ignore
        return



import logging
from typing import Union

__ALL__ = ["Connector"]

logger = logging.getLogger(__name__)


class Connector:
    """Connectors allow to explicitely make public a components scalar variable or variable group
    at the component's outer interface.

    An connector is for a component what an endpoint is for a connection.
    Connectors are hence the 'counterparts' to connection's endpoints.
    """

    def __init__(
        self,
        name: str,
        variable: Union[str, None] = None,
        variable_group: Union[str, None] = None,
        type: Union[str, None] = None,
    ):
        self.name: str = name
        self._variable: Union[str, None] = None
        self._variable_group: Union[str, None] = None
        self._type: Union[str, None] = None
        if variable:
            self.variable = variable
        if variable_group:
            self.variable_group = variable_group
        if type:
            self.type = type

    @property
    def variable(self) -> Union[str, None]:
        """Returns the scalar variable this connector is defined for.

        Returns
        -------
        Union[str, None]
            the scalar variable, if connector is a group connector. Otherwise None.
        """
        return self._variable

    @variable.setter
    def variable(self, variable: str):
        """Set the scalar variable this connector shall be defined for.

        Parameters
        ----------
        variable : str
            the scalar variable
        """
        if self._variable_group:
            msg = (
                f"Inconsistency: Connector {self.name} defines both variable and variableGroup.\n"
                f"variable: {variable}\nvariableGroup: {self._variable_group}\n"
                "variable is used. variableGroup is omitted."
            )
            logger.warning(msg)
            self._variable_group = None
        self._variable = variable

    @property
    def variable_group(self) -> Union[str, None]:
        """Returns the variable group this connector is defined for.

        Returns
        -------
        Union[str, None]
            the variable group, if connector is a group connector. Otherwise None.
        """
        return self._variable_group

    @variable_group.setter
    def variable_group(self, variable_group: str):
        """Set the variable group this connector shall be defined for.

        Parameters
        ----------
        variable_group : str
            the variable group
        """
        if self._variable:
            msg = (
                f"Inconsistency: Connector {self.name} defines both variable and variableGroup.\n"
                f"variable: {self._variable}\nvariableGroup: {variable_group}\n"
                "variable is omitted. variableGroup is used."
            )
            logger.warning(msg)
            self._variable = None
        self._variable_group = variable_group

    @property
    def type(self) -> Union[str, None]:
        """Returns the type of the connector."""
        return self._type

    @type.setter
    def type(self, type: str):
        """Set the type of the connector.

        Valid values are:
            "input"
            "output"
        """
        valid_types: list[str] = [
            "input",
            "output",
        ]
        if type not in valid_types:
            logger.error(f"connector {self.name}: type '{type}' is invalid.")
            return
        self._type = type
        return

    @property
    def is_single_connector(self) -> bool:
        """Returns True if connector is a single variable connector.

        Returns
        -------
        bool
            True if single variable connector. Otherwise False.
        """
        return bool(self._variable)

    @property
    def is_group_connector(self) -> bool:
        """Returns True if connector is a variable group connector.

        Returns
        -------
        bool
            True if variable group connector. Otherwise False.
        """
        return bool(self._variable_group)

    @property
    def variable_name(self) -> str:
        """Returns the name of the variable or variable group this connector is defined for.

        Returns
        -------
        str
            name of the variable or variable group
        """
        if self._variable:
            return self._variable
        elif self._variable_group:
            return self._variable_group
        else:
            return "UNKNOWN"
